__all__ = (
    'DefaultAvatar', 'FriendRequestFlag', 'HypesquadHouse', 'PremiumType', 'RelationshipType', 'Status', 'Theme'
)

from scarletio import copy_docs, export

from ...bases import Preinstance as P, PreinstancedBase
from ...color import Color
from ...http import urls as module_urls


class DefaultAvatar(PreinstancedBase, value_type = int):
    """
    Represents a default avatar of a user. Default avatar is used, when the user has no avatar set.
    
    There are some predefined default avatars and there should be no more instances created.
    
    Attributes
    ----------
    color : ``Color``
        The color of the default avatar.
    
    name : `str`
        The name of the default avatar's color.
    
    value : ``int`
        The identifier value of the default avatar.
    
    Type Attributes
    ---------------
    Every predefined instance can be accessed as a type attribute.
    
    +-----------+-----------+-----------+
    | name      | value     | color     |
    +===========+===========+===========+
    | blue      | 0         | 0x7289da  |
    +-----------+-----------+-----------+
    | gray      | 1         | 0x747f8d  |
    +-----------+-----------+-----------+
    | green     | 2         | 0x43b581  |
    +-----------+-----------+-----------+
    | orange    | 3         | 0xfaa61a  |
    +-----------+-----------+-----------+
    | red       | 4         | 0xf04747  |
    +-----------+-----------+-----------+
    | pink      | 5         | 0xff00a0  |
    +-----------+-----------+-----------+
    """
    __slots__ = ('color',)
    
    
    def __new__(cls, value, name = None, color = None):
        """
        Creates a default avatar.
        
        Parameters
        ----------
        value : `int`
            The identifier value of the default avatar.
        
        name : `None | str` = `None`, Optional
            The name of the default avatar's color.
        
        color : `None | Color` = `None`, Optional
            The color of the default avatar.
        """
        if color is None:
            color = Color()
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.color = color
        return self
    
    
    @classmethod
    def for_(cls, user):
        """
        Returns the default avatar for the given user.
        
        Parameters
        ----------
        user : ``UserBase``
            The user, who's default avatar will be looked up.

        Returns
        -------
        default_avatar : ``DefaultAvatar``
        """
        INSTANCES = cls.INSTANCES
        
        discriminator = user.discriminator
        if discriminator:
            key = discriminator
        else:
            key = user.id >> 22
        
        return INSTANCES[key % len(INSTANCES)]
    
    
    @copy_docs(PreinstancedBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        repr_parts.append(', color = ')
        repr_parts.append(repr(self.color))
    
    
    url = property(module_urls.default_avatar_url)
    
    
    # predefined
    blue = P(0, 'blue', Color(0x7289da))
    gray = P(1, 'gray', Color(0x747f8d))
    green = P(2, 'green', Color(0x43b581))
    orange = P(3, 'orange', Color(0xfaa61a))
    red = P(4, 'red', Color(0xf04747))
    pink = P(5, 'pink', Color(0xff00a0))


class HypesquadHouse(PreinstancedBase, value_type = int):
    """
    Represents Discord's hypesquad house.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the hypesquad house.
    
    name : `str`
        The name of the hypesquad house.
    
    Type Attributes
    ---------------
    Every predefined hypesquad house can also be accessed as type attribute:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | bravery               | bravery       | 1     |
    +-----------------------+---------------+-------+
    | brilliance            | brilliance    | 2     |
    +-----------------------+---------------+-------+
    | balance               | balance       | 3     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()

    # predefined
    none = P(0, 'none')
    bravery = P(1, 'bravery')
    brilliance = P(2, 'brilliance')
    balance = P(3, 'balance')


class Status(PreinstancedBase, value_type = str):
    """
    Represents a Discord user's status.
    
    Attributes
    ----------
    name : `str`
        The status's name.
    
    position : `int`
        Internal position of the status for sorting purposes.
    
    value : `str`
        The identifier value of the status.
    
    Type Attributes
    ---------------
    Each predefined status also can be accessed as a type attribute:
    
    +-----------------------+-----------+-----------+
    | Type attribute name   | position  | value     |
    +=======================+===========+===========+
    | online                | 0         | idle      |
    +-----------------------+-----------+-----------+
    | idle                  | 1         | idle      |
    +-----------------------+-----------+-----------+
    | dnd                   | 2         | dnd       |
    +-----------------------+-----------+-----------+
    | offline               | 3         | offline   |
    +-----------------------+-----------+-----------+
    | invisible             | 3         | invisible |
    +-----------------------+-----------+-----------+
    """
    __slots__ = ('position', )
    
    def __new__(cls, value, name = None, position = 4):
        """
        Creates a new status.
        
        Parameters
        ----------
        value : `str`
            The identifier value of the status.
        
        name : `None | str` = `None`, Optional
            The status's name.
        
        position : `int = `4`, Optional
            Internal position of the status for sorting purposes.
        """
        if name is None:
            name = value
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.position = position
        return self
    
    
    @copy_docs(PreinstancedBase.__lt__)
    def __lt__(self, other):
        if self is other:
            return False
        
        self_type = type(self)
        other_type = type(other)
        if self_type is other_type:
            pass
        
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return True
        
        else:
            return NotImplemented
        
        self_position = self.position
        other_position = other.position
        
        if self_position < other_position:
            return True
        
        if self_position > other_position:
            return False
        
        return (self.value < other.value)
    
    
    @copy_docs(PreinstancedBase.__gt__)
    def __gt__(self, other):
        if self is other:
            return False
        
        self_type = type(self)
        other_type = type(other)
        if self_type is other_type:
            pass
        
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return False
        
        else:
            return NotImplemented
        
        self_position = self.position
        other_position = other.position
        
        if self_position > other_position:
            return True
        
        if self_position < other_position:
            return False
        
        return (self.value > other.value)
    
    
    @copy_docs(PreinstancedBase.__eq__)
    def __eq__(self, other):
        if self is other:
            return True
        
        self_type = type(self)
        other_type = type(other)
        if self_type is other_type:
            pass
        
        elif issubclass(other_type, self_type.VALUE_TYPE):
            try:
                other = self_type.INSTANCES[other]
            except KeyError:
                return False
        
        else:
            return NotImplemented
    
        if self.position != other.position:
            return False
        
        if self.value != other.value:
            return False
        
        return True
    
    
    # predefined
    online = P('online', 'online', 0)
    idle = P('idle', 'idle', 1)
    dnd = P('dnd', 'dnd', 2)
    offline = P('offline', 'offline', 3)
    invisible = P('invisible', 'invisible', 3)


class PremiumType(PreinstancedBase, value_type = int):
    """
    Represents Discord's premium types.
    
    Attributes
    ----------
    name : `str`
        The default name of the premium type.
    
    value : `int`
        The Discord side identifier value of the premium type.
    
    Type Attributes
    ---------------
    Each predefined premium type can also be accessed as type attribute:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | name          | value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | nitro_classic         | nitro classic | 1     |
    +-----------------------+---------------+-------+
    | nitro                 | nitro         | 2     |
    +-----------------------+---------------+-------+
    | nitro_basic           | nitro basic   | 3     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    nitro_classic = P(1, 'nitro classic')
    nitro = P(2, 'nitro')
    nitro_basic = P(3, 'nitro basic')


@export
class RelationshipType(PreinstancedBase, value_type = int):
    """
    Represents a ``Relationship``'s type.
    
    Attributes
    ----------
    name : `str`
        The relationship type's name.
    
    value : `int`
        The Discord side identifier value of the relationship type.
    
    Type Attributes
    ---------------
    Each predefined relationship type can also be accessed as type attribute:
    
    +-----------------------+-------------------+-------+
    | Type attribute name   | name              | value |
    +=======================+===================+=======+
    | stranger              | stranger          | 0     |
    +-----------------------+-------------------+-------+
    | friend                | friend            | 1     |
    +-----------------------+-------------------+-------+
    | blocked               | blocked           | 2     |
    +-----------------------+-------------------+-------+
    | pending_incoming      | pending incoming  | 3     |
    +-----------------------+-------------------+-------+
    | pending_outgoing      | pending outgoing  | 4     |
    +-----------------------+-------------------+-------+
    | implicit              | implicit          | 5     |
    +-----------------------+-------------------+-------+
    | suggestion            | suggestion        | 6     |
    +-----------------------+-------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    stranger = P(0, 'stranger')
    friend = P(1, 'friend')
    blocked = P(2, 'blocked')
    pending_incoming = P(3, 'pending incoming')
    pending_outgoing = P(4, 'pending outgoing')
    implicit = P(5, 'implicit')
    suggestion = P(6, 'suggestion')


class FriendRequestFlag(PreinstancedBase, value_type = int):
    """
    Represents the friend request flags of a user.
    
    Attributes
    ----------
    name : `str`
        The default name of the friend request flag.
    
    value : `int`
        Internal identifier value of the friend request flag used for lookup.
    
    Type Attributes
    ---------------
    Every predefined friend request flag can also be accessed as type attribute:
    
    +---------------------------+---------------------------+-------+
    | Type attribute name       | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | mutual_guilds             | mutual_guilds             | 1     |
    +---------------------------+---------------------------+-------+
    | mutual_friends            | mutual_friends            | 2     |
    +---------------------------+---------------------------+-------+
    | mutual_guilds_and_friends | mutual_guilds_and_friends | 3     |
    +---------------------------+---------------------------+-------+
    | all                       | all                       | 4     |
    +---------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    
    @classmethod
    def decode(cls, data):
        """
        Converts the friend request flag data sent by Discord to it's wrapper side representation.
        
        Parameters
        ----------
        data : `None | dict<str, object>`
            Received friend request flag data.
        
        Returns
        -------
        friend_request_flag : `instance<type<cls>>`
        """
        if data is None:
            return cls.none
        
        all_ = data.get('all', False)
        if all_:
            key = 4
        else:
            mutual_guilds = data.get('mutual_guilds', False)
            mutual_friends = data.get('mutual_friends', False)
            
            key = mutual_guilds + (mutual_friends << 1)
        
        return cls.INSTANCES[key]
    
    
    def encode(self):
        """
        Returns the friend request flag's Discord side representation.
        
        Returns
        -------
        result : `dict` of (`str`, `bool`) items
        """
        value = self.value
        result = {}
        if (value >> 2) & 1:
            result['all'] = True
        else:
            if (value >> 1) & 1:
                result['mutual_friends'] = True
            if value & 1:
                result['mutual_guilds'] = True
        
        return result
    
    # predefined
    none = P(0, 'none')
    mutual_guilds = P(1, 'mutual guilds')
    mutual_friends = P(2, 'mutual friends')
    mutual_guilds_and_friends = P(3, 'mutual guilds and friends')
    all = P(4, 'all')


class Theme(PreinstancedBase, value_type = str):
    """
    Represents a user's theme.
    
    Attributes
    ----------
    name : `str`
        the theme's name.
    
    value : `str`
        The discord side identifier value of the theme.
    
    Type Attributes
    ---------------
    Each predefined theme instance can also be accessed as type attribute:
    
    +-----------------------+-----------+
    | Type attribute name   | value     |
    +=======================+===========+
    | dark                  | dark      |
    +-----------------------+-----------+
    | light                 | light     |
    +-----------------------+-----------+
    """
    __slots__ = ()
    
    # predefined
    dark = P('dark', 'dark')
    light = P('light', 'light')



__all__ = ()

from ....env import ALLOW_DEBUG_MESSAGES
from ....utils.debug import call_debug_logger

from ...field_parsers import entity_id_parser_factory, nullable_string_parser_factory, preinstanced_parser_factory
from ...field_putters import (
    entity_id_optional_putter_factory, entity_id_putter_factory, nullable_string_optional_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, nullable_entity_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)
from ...guild import Guild
from ...user import ClientUserBase

from ..audit_log_change import AuditLogChange
from ..audit_log_entry_change_conversion.value_mergers import value_merger_replace
from ..audit_log_change.flags import FLAG_HAS_AFTER, FLAG_HAS_BEFORE

from .constants import REASON_LENGTH_MAX, REASON_LENGTH_MIN
from .preinstanced import AuditLogEntryType


# changes


def parse_changes(data, entry_type = AuditLogEntryType.none):
    """
    Parses the entry's changes.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    entry_type : `None`, ``AuditLogEntryType`` = `AuditLogEntryType.none`, Optional
        The event's type.
    
    Returns
    -------
    changes : `None | dict<str, AuditLogChange>` 
    """
    change_datas = data.get('changes', None)
    if (change_datas is None) or (not change_datas):
        return None
    
    changes = None
    
    conversion_group = entry_type.target_type.change_conversions

    for change_data in change_datas:
        key = change_data.get('key', None)
        if key is None:
            continue
        
        if conversion_group is None:
            conversion = None
        else:
            conversion = conversion_group.get_conversion_for_key(key)
        
        if conversion is None:
            if ALLOW_DEBUG_MESSAGES:
                call_debug_logger(
                    (
                        f'Unknown audit log entry change key: {key!r}\n'
                        f'- Change data: {change_data!r}\n'
                        f'- Event type: {entry_type!r}'
                    ),
                    True,
                )
            continue
        
        for change in conversion.change_deserializer(conversion, change_data):
            if changes is None:
                changes = {}
            
            attribute_name = change.attribute_name
            current_change = changes.setdefault(attribute_name, change)
            if (change is not current_change):
                value_merger = conversion.value_merger
                if value_merger is None:
                    call_debug_logger(
                        (
                            f'Unexpected audit log entry merging (no or mismatching rules):\n'
                            f'- Instance 0: {change!r}\n'
                            f'- Instance 1: {current_change!r}'
                        ),
                        True,
                    )
                    value_merger = value_merger_replace
                
                changes[attribute_name] = AuditLogChange.from_fields(
                    attribute_name,
                    change.flags | current_change.flags,
                    value_merger(current_change.before, change.before),
                    value_merger(current_change.after, change.after),
                )

    return changes


def put_changes(changes, data, defaults, *, entry_type = AuditLogEntryType.none):
    """
    Serialises the given changes.
    
    Parameters
    ----------
    changes : `None | dict<str, AuditLogChange>` 
        Changes to serialize.
    data : `dict<str, object>`
        Data to extend.
    defaults : `bool`
        Whether fields with the default value should be ignored.
    entry_type : `None`, ``AuditLogEntryType`` = `AuditLogEntryType.none`, Optional (Keyword only)
        The event's type.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if changes is None:
        if defaults:
            data['changes'] = []
        return data
    
    change_datas = []
    
    conversion_group = entry_type.target_type.change_conversions
    
    for change in changes.values():
        if conversion_group is None:
            conversion = None
        else:
            conversion = conversion_group.get_conversion_for_name(change.attribute_name)
        
        if conversion is None:
            continue
        
        for change_data in conversion.change_serializer(conversion, change):
            change_datas.append(change_data)
    
    
    data['changes'] = change_datas
    return data


def validate_changes(changes, *, entry_type = AuditLogEntryType.none):
    """
    Validates the given changes.
    
    Parameters
    ----------
    changes : `None | iterable<AuditLogChange>`
        Changes to validate
    entry_type : `None`, ``AuditLogEntryType`` = `AuditLogEntryType.none`, Optional (Keyword only)
        The event's type.
    
    Returns
    -------
    changes : `None | dict<str, AuditLogChange>`
    
    Raises
    ------
    TypeError
        . If `changes` type is incorrect.
    ValueError
        - If `changes` value is incorrect.
    """
    if changes is None:
        return None
    
    if (getattr(type(changes), '__iter__', None) is None):
        raise TypeError(
            f'`changes` can be `None`, `iterable` of `{AuditLogChange.__name__}`, '
            f'got {type(changes).__name__}; {AuditLogChange!r}.'
        )
    
    
    conversion_group = entry_type.target_type.change_conversions
    
    validated_changes = None
    
    for change in changes:
        if not isinstance(change, AuditLogChange):
            raise TypeError(
                f'`changes` can contain `{AuditLogChange.__name__}` elements, '
                f'got {type(change).__name__}; {change!r}; changes = {changes!r}.'
            )
        
        attribute_name = change.attribute_name
        flags = change.flags
        validated_change = AuditLogChange.create_clean(attribute_name)
        
        if conversion_group is None:
            conversion = None
        else:
            conversion = conversion_group.get_conversion_for_name(attribute_name)
        
        if conversion is None:
            raise ValueError(
                f'No validator exists for `{attribute_name}`, got change = {change!r}; changes = {changes!r}.'
            )
        
        value_validator = conversion.value_validator
        
        if flags & FLAG_HAS_BEFORE:
            value = change.before
            if (value_validator is not None):
                value = value_validator(value)
            validated_change.before = value
            validated_change.flags |= FLAG_HAS_BEFORE
        
        if flags & FLAG_HAS_AFTER:
            value = change.after
            if (value_validator is not None):
                value = value_validator(value)
            validated_change.after = value
            validated_change.flags |= FLAG_HAS_AFTER
    
        
        if validated_changes is None:
            validated_changes = {}
        validated_changes[attribute_name] = validated_change
    
    return validated_changes


# details

def parse_details(data, entry_type = AuditLogEntryType.none):
    """
    Parses the entry's details.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    entry_type : `None`, ``AuditLogEntryType`` = `AuditLogEntryType.none`, Optional
        The event's type.
    
    Returns
    -------
    changes : `None | dict<str, object>` 
    """
    details_raw = data.get('options', None)
    if (details_raw is None) or (not details_raw):
        return None
    
    details = None
    
    conversion_group = entry_type.target_type.detail_conversions
    
    for key, value in details_raw.items():
        if conversion_group is None:
            conversion = None
        else:
            conversion = conversion_group.get_conversion_for_key(key)
        
        if conversion is None:
            if ALLOW_DEBUG_MESSAGES:
                call_debug_logger(
                    (
                        f'Unknown audit log entry detail key: {key!r}\n'
                        f'- Detail value: {value!r}\n'
                        f'- Event type: {entry_type!r}'
                    ),
                    True,
                )
            continue
        
        value_deserializer = conversion.value_deserializer
        if (value_deserializer is not None):
            value = value_deserializer(value)
        
        if details is None:
            details = {}
        
        details[conversion.field_name] = value
    
    return details


def put_details(details, data, defaults, *, entry_type = AuditLogEntryType.none):
    """
    Serialises the given details.
    
    Parameters
    ----------
    details : `None | dict<str, object>` 
        Changes to serialize.
    data : `dict<str, object>`
        Data to extend.
    defaults : `bool`
        Whether fields with the default value should be ignored.
    entry_type : `None`, ``AuditLogEntryType`` = `AuditLogEntryType.none`, Optional (Keyword only)
        The event's type.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if details is None:
        if defaults:
            data['options'] = []
        return data
    
    conversion_group = entry_type.target_type.detail_conversions
    
    details_raw = {}
    
    for name, value in details.items():
        if conversion_group is None:
            conversion = None
        else:
            conversion = conversion_group.get_conversion_for_name(name)
        if conversion is None:
            continue
        
        value_serializer = conversion.value_serializer
        if (value_serializer is not None):
            value = value_serializer(value)
        
        details_raw[conversion.field_key] = value
    
    data['options'] = details_raw
    return data


def validate_details(details, *, entry_type = AuditLogEntryType.none):
    """
    Validates the given details.
    
    Parameters
    ----------
    details : `None | dict<str, object>`
        Changes to validate
    entry_type : `None`, ``AuditLogEntryType`` = `AuditLogEntryType.none`, Optional (Keyword only)
        The event's type.
    
    Returns
    -------
    details : `None | dict<str, object>`
    
    Raises
    ------
    TypeError
        . If `details` type is incorrect.
    ValueError
        - If `details` value is incorrect.
    """
    if details is None:
        return None
    
    if not isinstance(details, dict):
        raise TypeError(
            f'`details` can be `None`, dict`, got {type(details).__name__}; {details!r}.'
        )
    
    conversion_group = entry_type.target_type.detail_conversions
    
    validated_details = None
    
    for name, value in details.items():
        if not isinstance(name, str):
            raise TypeError(
                f'`details` keys can be `str`, got {type(name).__name__}; {name!r}; details = {details!r}.'
            )
        
        if conversion_group is None:
            conversion = None
        else:
            conversion = conversion_group.get_conversion_for_name(name)
        if conversion is None:
            raise ValueError(
                f'No validator for `{name}` of exists, got value = {value!r}; details = {details!r}.'
            )
        
        value_validator = conversion.value_validator
        if (value_validator is not None):
            value = value_validator(value)
        
        if validated_details is None:
            validated_details = {}
        validated_details[name] = value
    
    return validated_details


# guild_id

def parse_guild_id(data, parent = None):
    """
    Parses out the entry's guild's identifier.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    parent : `None`, ``AuditLog`` = `None`, Optional
        The parent of the entry.
    
    Returns
    -------
    guild_id : `int`
    """
    if parent is None:
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
    else:
        guild_id = parent.guild_id
    
    return guild_id


put_guild_id = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('entry_id')

# parent

validate_parent = nullable_entity_validator_factory('parent', NotImplemented, include = 'AuditLog')

# reason

parse_reason = nullable_string_parser_factory('reason')
put_reason = nullable_string_optional_putter_factory('reason')
validate_reason = nullable_string_validator_factory('reason', REASON_LENGTH_MIN, REASON_LENGTH_MAX)

# target_id

parse_target_id = entity_id_parser_factory('target_id')
put_target_id = entity_id_optional_putter_factory('target_id')
validate_target_id = entity_id_validator_factory('target_id')

# type

parse_type = preinstanced_parser_factory('action_type', AuditLogEntryType, AuditLogEntryType.none)
put_type = preinstanced_putter_factory('action_type')
validate_type = preinstanced_validator_factory('entry_type', AuditLogEntryType)

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_optional_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ('EmbedProvider',)

from scarletio import copy_docs

from ...utils import sanitize_mentions

from ..embed_field_base import EmbedFieldBase

from .fields import parse_name, parse_url, put_name, put_url, validate_name, validate_url


class EmbedProvider(EmbedFieldBase):
    """
    Represents an embed's provider.
    
    Attributes
    ----------
    name : `None`, `str`
        The name of the provider.
    url : `None`, `str`
        The url of the provider.
    """
    __slots__ = ('name', 'url')
    
    def __new__(cls, name = ..., url = ...):
        """
        Creates a new embed field with the given parameters.
        
        Parameters
        ----------
        name : `None`, `str`, Optional
            The name of the provider.
        url : `None`, `str`, Optional
            The url of the provider.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # name
        if name is ...:
            name = None
        else:
            name = validate_name(name)
        
        # url
        if url is ...:
            url = None
        else:
            url = validate_url(url)
        
        self = object.__new__(cls)
        self.name = name
        self.url = url
        return self
    
        
    @copy_docs(EmbedFieldBase.__len__)
    def __len__(self):
        length = 0
        
        # name
        name = self.name
        if (name is not None):
            length += len(name)
        
        return length
    
    
    @copy_docs(EmbedFieldBase.__bool__)
    def __bool__(self):
        # name
        if self.name is not None:
            return True
        
        # url
        if self.url is not None:
            return True
        
        return False
    
    
    @copy_docs(EmbedFieldBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        field_added = False
        
        name = self.name
        if name is not None:
            field_added = True
            repr_parts.append(' name = ')
            repr_parts.append(repr(self.name))
        
        url = self.url
        if url is not None:
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' url = ')
            repr_parts.append(repr(url))
    
    
    @copy_docs(EmbedFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # name
        name = self.name
        if (name is not None):
            hash_value ^= hash(name)
        
        # url
        url = self.url
        if (url is not None):
            hash_value ^= hash(url)
        
        return hash_value
    
    
    @copy_docs(EmbedFieldBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # name
        if self.name != other.name:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(EmbedFieldBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.name = parse_name(data)
        self.url = parse_url(data)
        return self
    
    
    @copy_docs(EmbedFieldBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        
        put_name(self.name, data, defaults)
        put_url(self.url, data, defaults)
        
        return data
    
    
    @copy_docs(EmbedFieldBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        new.name = sanitize_mentions(self.name, guild)
        new.url = self.url
        return new
    
    
    @copy_docs(EmbedFieldBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.name = self.name
        new.url = self.url
        return new
    
    
    def copy_with(self, *, name = ..., url = ...):
        """
        Copies the embed provider with the given parameters.
        
        Parameters
        ----------
        name : `None`, `str, Optional (Keyword only)
            The name of the provider.
        url : `None`, `str`, Optional (Keyword only)
            The url of the provider.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # url
        if url is ...:
            url = self.url
        else:
            url = validate_url(url)
        
        new = object.__new__(type(self))
        new.name = name
        new.url = url
        return new
    
    
    @copy_docs(EmbedFieldBase.iter_contents)
    def iter_contents(self):
        name = self.name
        if (name is not None):
            yield name



from .onboarding_prompt import *
from .onboarding_prompt_option import *
from .onboarding_screen import *


__all__ = (
    *onboarding_prompt.__all__,
    *onboarding_prompt_option.__all__,
    *onboarding_screen.__all__,
)



__all__ = ('VoiceChannelEffectAnimationType',)

from ...bases import Preinstance as P, PreinstancedBase


class VoiceChannelEffectAnimationType(PreinstancedBase, value_type = int):
    """
    Voice channel effect animation type.
    
    Attributes
    ----------
    name : `str`
        The default name of the voice channel effect animation type.
    
    value : `int`
        The discord side identifier value of the voice channel effect animation type.
    
    Type Attributes
    ---------------
    Every predefined voice channel effect animation type can be accessed as type attribute as well:
    
    +-----------------------------------+---------------------------+-------+
    | Type attribute name               | Name                      | Value |
    +===================================+===========================+=======+
    | premium                           | premium                   | 0     |
    +-----------------------------------+---------------------------+-------+
    | basic                             | basic                     | 1     |
    +-----------------------------------+---------------------------+-------+
    """
    __slots__ = ()
    
    # predefined
    premium = P(0, 'premium')
    basic = P(1, 'basic')



__all__ = ('Reaction',)

from scarletio import RichAttributeErrorBaseType

from ..emoji import Emoji

from .fields import parse_emoji, parse_type, put_emoji, put_type, validate_emoji, validate_type
from .preinstanced import ReactionType


class Reaction(RichAttributeErrorBaseType):
    """
    Represents a reaction.
    
    Parameters
    ----------
    emoji : ``Emoji``
        The reaction emoji,
    reaction_type : ``ReactionType``
        The reaction's type
    """
    __slots__ = ('emoji', 'type')
    
    def __new__(cls, emoji, *, reaction_type = ...):
        """
        Creates a new new reaction with the given parameters.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The reaction emoji,
        reaction_type : ``ReactionType``, `int`, Optional (Keyword only)
            The reaction's type.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # emoji
        emoji = validate_emoji(emoji)
        
        # reaction_type
        if reaction_type is ...:
            reaction_type = ReactionType.standard
        else:
            reaction_type = validate_type(reaction_type)
        
        # Construct
        self = object.__new__(cls)
        self.emoji = emoji
        self.type = reaction_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new reaction.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Reaction data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        emoji = parse_emoji(data)
        reaction_type = parse_type(data)
        
        self = object.__new__(cls)
        self.emoji = emoji
        self.type = reaction_type
        return self
    
    
    @classmethod
    def from_fields(cls, emoji, reaction_type):
        """
        Creates a new reaction from the given fields. Not like ``.__new__``, this has no validation.
        
        Parameters
        ----------
        emoji : ``Emoji``
            The reaction emoji,
        reaction_type : ``ReactionType``
            The reaction's type
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.emoji = emoji
        self.type = reaction_type
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the reaction field.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_emoji(self.emoji, data, defaults)
        put_type(self.type, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the reaction's representation."""
        repr_parts = ['<',self.__class__.__name__]
        
        repr_parts.append(' emoji = ')
        repr_parts.append(repr(self.emoji))
        
        reaction_type = self.type
        repr_parts.append(', type = ')
        repr_parts.append(reaction_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(reaction_type.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two reaction's are equal."""
        other_type = type(other)
        if other_type is type(self):
            if self.emoji is not other.emoji:
                return False
            
            if self.type is not other.type:
                return False
            
            return True
        
        if other_type is Emoji:
            if self.emoji is not other:
                return False
            
            # If the other entity is an Emoji, only match it if we are standard
            if self.type is not ReactionType.standard:
                return False
            
            return True
        
        return NotImplemented
    
    
    def __hash__(self):
        """returns the reaction's hash value."""
        hash_value = 0
        
        # emoji
        hash_value ^= hash(self.emoji)
        
        # reaction_type
        hash_value ^= hash(self.type)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the reaction returning a new one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.emoji = self.emoji
        new.type = self.type
        return new
    
    
    def copy_with(self, *, emoji = ..., reaction_type = ...):
        """
        Copies the reaction with the given fields.
        
        Parameters
        ----------
        emoji : ``Emoji``, Optional (Keyword only)
            The reaction emoji,
        reaction_type : ``ReactionType``, Optional (Keyword only)
            The reaction's type.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # type
        if reaction_type is ...:
            reaction_type = self.type
        else:
            reaction_type = validate_type(reaction_type)
        
        # Construct
        new = object.__new__(type(self))
        new.emoji = emoji
        new.type = reaction_type
        return new



__all__ = ('Client', )

import sys, warnings
from json import JSONDecodeError
from math import inf

from scarletio import (
    CancelledError, CompoundMetaType, EventThread, Future, LOOP_TIME, Task, copy_docs, export, from_json,
    methodize, run_coroutine, sleep, write_exception_async
)

from ...env import CACHE_USER
from ...ext import get_setup_functions, run_setup_functions

from ..activity import ACTIVITY_UNKNOWN
from ..application import Application, Team
from ..core import APPLICATION_ID_TO_CLIENT, CHANNELS, CLIENTS, GUILDS, KOKORO, USERS
from ..events import IntentFlag
from ..events.core import register_client, unregister_client
from ..events.event_handler_manager import EventHandlerManager
from ..events.handling_helpers import ensure_shutdown_event_handlers, ensure_voice_client_shutdown_event_handlers
from ..exceptions import (
    DiscordException, DiscordGatewayException, INTENT_ERROR_CODES, InvalidToken, RESHARD_ERROR_CODES
)
from ..gateway.utils import DiscordGatewayClientBase, create_gateway, reshard_gateway
from ..http import DiscordApiClient, RateLimitProxy
from ..localization.utils import LOCALE_DEFAULT
from ..user import (
    ClientUserBase, ClientUserPBase, PremiumType, RelationshipType, Status, User, UserBase, UserFlag,
    create_partial_user_from_id
)
from ..user.user.fields import (
    parse_email, parse_email_verified, parse_locale, parse_mfa_enabled, parse_premium_type, validate_avatar_decoration,
    validate_banner_color, validate_bot, validate_clan, validate_discriminator, validate_display_name, validate_email,
    validate_email_verified, validate_flags, validate_locale, validate_mfa_enabled, validate_name,
    validate_premium_type, validate_status
)

from .compounds import CLIENT_COMPOUNDS
from .fields import (
    validate_activity, validate_additional_owner_ids, validate_api, validate_application_id, validate_client_id,
    validate_extensions, validate_http, validate_http_debug_options, validate_intents, validate_secret,
    validate_shard_count, validate_should_request_users, validate_token
)
from .functionality_helpers import _check_is_client_duped, try_get_user_id_from_token
from .ready_state import ReadyState


AUTO_CLIENT_ID_LIMIT = 1 << 22


@export
class Client(
    ClientUserPBase,
    *CLIENT_COMPOUNDS,
    build = True,
    metaclass = CompoundMetaType.with_(ClientUserPBase),
):
    """
    Discord client class used to interact with the Discord API.
    
    Attributes
    ----------
    _activity : ``Activity``
        The client's preferred activity.
    
    _additional_owner_ids : `None`, `set` of `int`
        Additional users' (as id) to be passed by the ``.is_owner`` check.
    
    _gateway_max_concurrency : `int`
        The maximal amount of shards that can be launched at the same time.
    
    _gateway_requesting : `bool`
        Whether the client already requests it's gateway.
    
    _gateway_time : `float`
        The last timestamp when ``._gateway_url`` was updated.
    
    _gateway_url : `str`
        Cached up gateway url, what is invalidated after `1` minute. Used to avoid unnecessary requests when launching
        up more shards.
    
    _gateway_waiter : `None`, ``Future``
        When client gateway is being requested multiple times at the same time, this future is set and awaited at the
        secondary requests.
    
    _should_request_users : `bool`
        Whether the client should try to request the users of it's guilds.
    
    _status : ``Status``
        The client's preferred status.
    
    _user_chunker_nonce : `int`
        The last nonce in int used for requesting guild user chunks. The default value is `0`, what means the next
        request will start at `1`.
        
        Nonce `0` is allocated for the case, when all the guild's users are requested.
    
    activities : `None`, `list` of ``Activity``
        A list of the client's activities. Defaults to `None`.
    
    api : ``DiscordApiClient``
        Discord api client for lower level interactions with the Discord API.
    
    application : ``Application``
        The client's application. The application data of the client is requested meanwhile it logs in.
    
    avatar_hash : `int`
        The client's avatar's hash in `uint128`.
    
    avatar_type : ``IconType``
        The client's avatar's type.
    
    avatar_decoration : `None`, ``AvatarDecoration``
        The client's avatar decorations.
    
    banner_color : `None`, ``Color``
        The client's banner color if has any.
    
    banner_hash : `int`
        The client's banner's hash in `uint128`.
    
    banner_type : ``IconType``
        The client's banner's type.
    
    bot : `bool`
        Whether the client is a bot.
    
    clan : `None`, ``UserClan``
        The client's primary clan.
    
    discriminator : `int`
        The client's discriminator. Given to avoid overlapping names.
    
    display_name : `None`, `str`
        The clients' non-unique display name.
    
    email : `None`, `str`
        The client's email.
    
    email_verified : `bool`
        Whether the email of the client is verified.
    
    events : ``EventHandlerManager``
        Contains the event handlers of the client. New event handlers can be added through it as well.
    
    flags : ``UserFlag``
        The client's user flags.
    
    gateway : ``DiscordGatewayClientBase``
        The gateway of the client towards Discord.
    
    group_channels : `dict` of (`int`, ``Channel``) items
        The group channels of the client. They can be accessed by their id as the key.
    
    guild_profiles : `dict` of (`int`, ``GuildProfile``) items
        A dictionary, which contains the client's guild profiles. If a client is member of a guild, then it should
        have a respective guild profile accordingly.
    
    guilds : `set` of ``Guild``
        The guilds, where the client is in.
    
    http : ``HttpClient``
        The http session of the client.
    
    id : `int`
        The client's unique identifier number.
    
    intents : ``IntentFlag``
        The intent flags of the client.
    
    locale : ``Locale``
        The preferred locale by the client.
    
    mfa_enabled : `bool`
        Whether the client has two factor authorization enabled on the account.
    
    name : str
        The client's username.
    
    premium_type : ``PremiumType``
        The Nitro subscription type of the client.
    
    private_channels : `dict` of (`int`, ``Channel``) items
        Stores the private channels of the client. The channels' other recipient' ids are the keys, meanwhile the
        channels are the values.
    
    ready_state : `None`, ``ReadyState``
        The client on login fills up it's ``.ready_state`` with ``Guild`` objects, which will have their members
        requested.
        
        When receiving a `READY` dispatch event, the client's ``.ready_state`` is set as a ``ReadyState`` and
        a ``._delay_ready`` task is started, what delays the handle-able `ready` event, till every user from the
        received guilds is cached up. When done, ``.ready_state`` is set back to `None`.
    
    relationships : `dict` of (`int`, ``Relationship``) items
        Stores the relationships of the client. The relationships' users' ids are the keys and the relationships
        themselves are the values.
    
    running : `bool`
        Whether the client is running or not. When the client is stopped, this attribute is set as `False` what causes
        it's heartbeats to stop and it's gateways to close and not reconnect.
    
    secret : `str`
        The client's secret used when interacting with oauth2 endpoints.
    
    shard_count : `int`
        The client's shard count. Set as `0` if the bot is not using sharding.
    
    status : `Status`
        The client's display status.
    
    statuses : `None`, `dict` of (`str`, `str`) items
        The client's statuses for each platform.
    
    thread_profiles : `None`, `dict` (``Channel``, ``ThreadProfile``) items
        A Dictionary which contains the thread profiles for the user in thread channel - thread profile relation.
        Defaults to `None`.
    
    token : `str`
        The client's token.
    
    voice_clients : `dict` of (`int`, ``VoiceClient``) items
        Each bot can join a channel at every ``Guild`` and meanwhile they do, they have an active voice client for that
        guild. This attribute stores these voice clients. They keys are the guilds' ids, meanwhile the values are
        the voice clients.
    
    
    Class Attributes
    ----------------
    loop : ``EventThread``
        The event loop of the client. Every client uses the same one.
    _next_auto_id : `int`
        Auto id generator for clients without identifier.
    
    See Also
    --------
    - ``UserBase`` : The superclass of ``Client`` and of other user types.
    - ``User`` : The default type of Discord users.
    - ``Webhook`` : Discord webhook entity.
    - ``WebhookRepr`` : Discord webhook's user representation.
    - ``Oauth2User`` : A user class with extended oauth2 attributes.
    
    Notes
    -----
    Client supports weakreferencing and dynamic attribute names as well for extension support.
    """
    __slots__ = (
        '__dict__', '_activity', '_additional_owner_ids', '_gateway_max_concurrency', '_gateway_requesting',
        '_gateway_time', '_gateway_url', '_gateway_waiter', '_should_request_users', '_status', '_user_chunker_nonce',
        'api', 'application', 'email', 'email_verified', 'events', 'gateway', 'group_channels', 'guilds', 'http',
        'intents', 'locale', 'mfa_enabled', 'premium_type', 'private_channels', 'ready_state', 'relationships',
        'running', 'secret', 'shard_count', 'token', 'voice_clients'
    )
    
    loop = KOKORO
    _next_auto_id = 1
    
    def __new__(
        cls,
        token,
        *,
        activity = ...,
        additional_owners = ...,
        api = ...,
        application_id = ...,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        bot = ...,
        clan = ...,
        client_id = ...,
        discriminator = ...,
        display_name = ...,
        email = ...,
        email_verified = ...,
        extensions = ...,
        flags = ...,
        http = ...,
        http_debug_options = ...,
        intents = ...,
        locale = ...,
        mfa_enabled = ...,
        name = ...,
        premium_type = ...,
        secret = ...,
        shard_count = ...,
        should_request_users = ...,
        status = ...,
        **keyword_parameters,
    ):
        """
        Creates a new ``Client`` with the given parameters.
        
        Parameters
        ----------
        token : `str`
            A valid Discord token, what the client can use to interact with the Discord API.
        
        activity : ``Activity``, Optional (Keyword only)
            The client's preferred activity.
        
        additional_owners : `None`, `int`, ``ClientUserBase``, `iterable` of (`int`, ``ClientUserBase``) \
                , Optional (Keyword only)
            Additional users to return `True` for by ``.is_owner`.
        
        api : `None | DiscordApiClient`, Optional (Keyword only)
            The api client to use.
        
        application_id : `None`, `int`, `str`, Optional (Keyword only)
            The client's application id. If passed as `str`, will be converted to `int`.
         
        avatar : `None`, ``Icon``, `str`, Optional (Keyword only)
            The client's avatar.
        
        avatar_decoration : `None`, ``AvatarDecoration``, Optional (Keyword only)
            The client's avatar decoration.
        
        banner : `None`, ``Icon``, `str`, Optional (Keyword only)
            The client's banner.
        
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The client's banner color.
        
        bot : `bool`, Optional (Keyword only)
            Whether the client is a bot.
        
        clan : `None`, ``UserClan``, Optional (Keyword only)
            The client's primary clan.
        
        client_id : `None`, `int`, `str`, Optional (Keyword only)
            The client's `.id`. If passed as `str` will be converted to `int`. Defaults to `None`.
            
            When more `Client` is started up, it is recommended to define their id initially. The wrapper can detect the
            clients' id-s only when they are logging in, so the wrapper  needs to check if a ``User`` alter_ego of the
            client exists anywhere, and if does will replace it.
        
        discriminator : `str`, `int`, Optional (Keyword only)
            The client's discriminator.
        
        display_name : `None`, `str`, Optional (Keyword only)
            The client's non-unique display name.
        
        email : `None, `str`, Optional (Keyword only)
            The client's email.
        
        email_verified : `bool`, Optional (Keyword only)
            Whether the email of the client is verified.
        
        extensions : `None`, `str`, `iterable` of `str`, Optional (Keyword only)
            The extension's name to setup on the client.
        
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        
        http : `None | HTTPClient`, Optional (Keyword only)
            The http client to use.
        
        http_debug_options: `None`, `str`, `iterable` of `str`, Optional (Keyword only)
            Http client debug options for the client.
        
        intents : `int`, ``IntentFlag``, Optional (Keyword only)
             By default the client will launch up using all the intent flags. Negative values will be interpreted as
             using all the intents, meanwhile if passed as positive, non existing intent flags are removed.
        
        locale : ``Locale``, `str`, Optional (Keyword only)
            The preferred locale by the client.
        
        mfa_enabled : `bool`, Optional (Keyword only)
            Whether the user has two factor authorization enabled on the account.
        
        name : `str`, Optional (Keyword only)
            The user's name.
        
        premium_type : ``PremiumType``, `int`, Optional (Keyword only)
            The Nitro subscription type of the client.
        
        secret: `str`, Optional (Keyword only)
            Client secret used when interacting with oauth2 endpoints.
        
        shard_count : `int`, Optional (Keyword only)
            The client's shard count. If passed as lower as the recommended one, will reshard itself.
        
        should_request_users : `int`, Optional (Keyword only)
            Whether the client should try to request the users of it's guilds.
        
        status : `None`, `str`, ``Status``, Optional (Keyword only)
            The client's preferred status.
        
        **keyword_parameters : keyword parameters
            Additional parameters to pass to extension setup functions.
            
            > If any required parameter by an extension is missing `RuntimeError` is raised, meanwhile if any extra
            > is given, `RuntimeWarning` is dropped.
        
        
        Returns
        -------
        client : ``Client``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        RuntimeError
            Creating the same client multiple times is not allowed.
        """
        # ---- Required Parameters ----
        
        # token
        token = validate_token(token)
        
        # ---- Optional Client Parameters ----
        
        # activity
        if activity is ...:
            activity = ACTIVITY_UNKNOWN
        else:
            activity = validate_activity(activity)
        
        # additional owners
        if additional_owners is ...:
            additional_owner_ids = None
        else:
            additional_owner_ids = validate_additional_owner_ids(additional_owners)
        
        # api
        if api is ...:
            api = None
        else:
            api = validate_api(api)
        
        # application_id
        if (application_id is ...):
            application_id = 0
        else:
            application_id = validate_application_id(application_id)
        
        application = Application._create_empty(application_id)
        
        # avatar
        if avatar is ...:
            avatar = None
        else:
            avatar = cls.avatar.validate_icon(avatar)
        
        # avatar_decoration
        if avatar_decoration is ...:
            avatar_decoration = None
        else:
            avatar_decoration = validate_avatar_decoration(avatar_decoration)
        
        # banner
        if banner is ...:
            banner = None
        else:
            banner = cls.banner.validate_icon(banner)
        
        # banner_color
        if banner_color is ...:
            banner_color = None
        else:
            banner_color = validate_banner_color(banner_color)
        
        # bot
        if bot is ...:
            bot = True
        else:
            bot = validate_bot(bot)
        
        # clan
        if clan is ...:
            clan = None
        else:
            clan = validate_clan(clan)
        
        # client_id
        if client_id is ...:
            client_id = try_get_user_id_from_token(token)
        else:
            client_id = validate_client_id(client_id)
        
        # discriminator
        if discriminator is ...:
            discriminator = 0
        else:
            discriminator = validate_discriminator(discriminator)
        
        # display_name
        if display_name is ...:
            display_name = None
        else:
            display_name = validate_display_name(display_name)
        
        # email
        if email is ...:
            email = None
        else:
            email = validate_email(email)
        
        # email_verified
        if email_verified is ...:
            email_verified = False
        else:
            email_verified = validate_email_verified(email_verified)
        
        # extensions
        if extensions is ...:
            extensions = None
        else:
            extensions = validate_extensions(extensions)
        
        # flags
        if flags is ...:
            flags = UserFlag()
        else:
            flags = validate_flags(flags)
        
        # http
        if http is ...:
            http = None
        else:
            http = validate_http(http)
        
        # http_debug_options
        if http_debug_options is ...:
            http_debug_options = None
        else:
            http_debug_options = validate_http_debug_options(http_debug_options)
        
        # intents
        if intents is ...:
            intents = IntentFlag(-1)
        else:
            intents = validate_intents(intents)
        
        # locale
        if locale is ...:
            locale = LOCALE_DEFAULT
        else:
            locale = validate_locale(locale)
        
        # mfa_enabled
        if mfa_enabled is ...:
            mfa_enabled = False
        else:
            mfa_enabled = validate_mfa_enabled(mfa_enabled)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # premium_type
        if premium_type is ...:
            premium_type = PremiumType.none
        else:
            premium_type = validate_premium_type(premium_type)
        
        # secret
        if secret is ...:
            secret = ''
        else:
            secret = validate_secret(secret)
        
        # shard count
        if shard_count is ...:
            shard_count = 0
        else:
            shard_count = validate_shard_count(shard_count)
        
        # should_request_users
        if should_request_users is ...:
            should_request_users = True
        else:
            should_request_users = validate_should_request_users(should_request_users)
        
        # status
        if status is ...:
            status = Status.online
        else:
            status = validate_status(status)
            if status is Status.offline:
                status = Status.invisible
        
        # ---- Setup extensions ----
        
        # extensions
        setup_functions = get_setup_functions(extensions, keyword_parameters)
        
        
        # ---- Auto generate initial id if un-detected ----
        
        if client_id < AUTO_CLIENT_ID_LIMIT:
            client_id = cls._next_auto_id
            cls._next_auto_id = client_id + 1
        
        # ---- Api client ----
        
        if api is None:
            api = DiscordApiClient(bot, token, debug_options = http_debug_options, http = http)
        
        http = api.http
        
        # ---- Build object ----
        
        # Set all Attributes
        
        self = object.__new__(cls)
        # Set id & name first, so if an exception occurs `repr(self)` wont fail.
        self.id = client_id
        self.name = name
        
        self._activity = activity
        self._additional_owner_ids = additional_owner_ids
        self._gateway_max_concurrency = 1
        self._gateway_requesting = False
        self._gateway_time = -inf
        self._gateway_url = ''
        self._gateway_waiter = None
        self._should_request_users = should_request_users
        self._status = status
        self._user_chunker_nonce = 0
        
        self.activities = None
        self.api = api
        self.application = application
        self.avatar = avatar
        self.avatar_decoration = avatar_decoration
        self.banner = banner
        self.banner_color = banner_color
        self.bot = bot
        self.clan = clan
        self.discriminator = discriminator
        self.display_name = display_name
        self.email = email
        self.email_verified = email_verified
        self.events = EventHandlerManager(self)
        self.flags = flags
        self.gateway = DiscordGatewayClientBase()
        self.group_channels = {}
        self.guild_profiles = {}
        self.guilds = set()
        self.http = http
        self.intents = intents
        self.locale = locale
        self.mfa_enabled = mfa_enabled
        self.premium_type = premium_type
        self.private_channels = {}
        self.ready_state = None
        self.relationships = {}
        self.running = False
        self.secret = secret
        self.shard_count = shard_count
        self.status = Status.offline
        self.statuses = None
        self.thread_profiles = None
        self.token = token
        self.voice_clients = {}
        
        # These might require other attributes to be set
        self.gateway = create_gateway(self)
        
        # Check whether the client is duped
        if client_id > AUTO_CLIENT_ID_LIMIT:
            _check_is_client_duped(self, client_id)
            self._maybe_replace_alter_ego()
        
        # Setup extensions
        run_setup_functions(self, setup_functions, keyword_parameters)
        
        # Register client
        CLIENTS[client_id] = self
        
        # Register client by application id
        if application_id:
            APPLICATION_ID_TO_CLIENT[application_id] = self
        
        
        return self
    
    
    def _maybe_replace_alter_ego(self):
        """
        Replaces the type ``User`` alter_ego of the client if applicable.
        """
        client_id = self.id
        
        # GOTO
        while True:
            if CACHE_USER:
                try:
                    alter_ego = USERS[client_id]
                except KeyError:
                    # Go Out
                    break
                else:
                    if alter_ego is not self:
                        # we already exists, we need to go tru everything and replace our self.
                        guild_profiles = alter_ego.guild_profiles
                        self.guild_profiles = guild_profiles
                        for guild_id in guild_profiles.keys():
                            try:
                                guild = GUILDS[guild_id]
                            except KeyError:
                                continue
                            
                            guild.users[client_id] = self
            
            # This part should run at both case, except when there is no alter_ego detected when caching users.
            for client in CLIENTS.values():
                if (client is self) or (not client.running):
                    continue
                
                for channel in client.group_channels.values():
                    users = channel.users
                    for index in range(len(users)):
                        if users[index].id == client_id:
                            users[index] = self
                            break
                
                for channel in client.private_channels.values():
                    users = channel.users
                    for index in range(len(users)):
                        if users[index].id == client_id:
                            users[index] = self
                            break
            
            break
        
        USERS[client_id] = self
    
    
    def _init_on_ready(self, data):
        """
        Fills up the client's instance attributes on login. If there is an already existing User object with the same
        id, the client will replace it at channel participants, at ``USERS`` weakreference dictionary, at
        ``guild.users``. This replacing is avoidable, if at the creation of the client the ``.client_id`` parameter is
        set.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Data requested from Discord by the ``.client_login_static`` method.
        
        Raises
        ------
        RuntimeError
            Creating the same client multiple times is not allowed.
        """
        client_id = int(data['id'])
        _check_is_client_duped(self, client_id)
        
        self_id = self.id
        if self_id != client_id:
            if CLIENTS.get(self_id, None) is self:
                del CLIENTS[self_id]
            
            if USERS.get(self_id, None) is self:
                del USERS[self_id]
            
            CLIENTS[client_id] = self
            self.id = client_id
        
        self._maybe_replace_alter_ego()
        self._update_attributes(data)
    
    
    @property
    def _platform(self):
        """
        Returns the client's local platform.
        
        Returns
        -------
        platform : `str`
            The platform's name or empty string if the client's status is offline or invisible.
            
        Notes
        -----
        Custom client's status is always `'web'`, so other than `''`, `'web'` will not be returned.
        """
        if self.status in (Status.offline, Status.invisible):
            return ''
        
        return 'web'
    
    
    def _delete(self):
        """
        Clears up the client's references. By default this is not called when a client is stopped. This method should
        be used when you want to get rid of every allocated objects by the client. Take care, local modules might still
        have active references to the client or to some other objects, what could cause them to not garbage collect.
        
        Raises
        ------
        RuntimeError
            If called when the client is still running.
        
        Examples
        --------
        ```py
        >>> from hata import Client, GUILDS
        >>> import gc
        >>> TOKEN = 'a token goes here'
        >>> client = Client(TOKEN)
        >>> client.start()
        >>> len(GUILDS)
        4
        >>> client.stop()
        >>> client._delete()
        >>> client = None
        >>> gc.collect()
        680
        >>> len(GUILDS)
        0
        ```
        """
        if self.running:
            raise RuntimeError(f'`{type(self).__name__}._delete` called from a running client: {self!r}')
        
        client_id = self.id
        if CLIENTS.get(client_id, None) is self:
            del CLIENTS[client_id]
        
        
        application_id = self.application.id
        if APPLICATION_ID_TO_CLIENT.get(application_id, None) is self:
            del APPLICATION_ID_TO_CLIENT[application_id]
        
        if USERS.get(client_id, None) is self:
            alter_ego = User._from_client(self, True)
            USERS[client_id] = alter_ego
            
            guild_profiles = self.guild_profiles
            for guild_id in guild_profiles.keys():
                try:
                    guild = GUILDS[guild_id]
                except KeyError:
                    continue
                
                guild.users[client_id] = alter_ego
            
            for client in CLIENTS.values():
                if (client is not self) and client.running:
                    for relationship in client.relationships:
                        if relationship.user is alter_ego:
                            relationship.user = alter_ego
            
            thread_profiles = self.thread_profiles
            if (thread_profiles is not None):
                for channel_id in thread_profiles.keys():
                    try:
                        channel = CHANNELS[channel_id]
                    except KeyError:
                        pass
                    else:
                        thread_users = channel.thread_users
                        if (thread_users is not None):
                            thread_users[client_id] = alter_ego
            
            self.relationships.clear()
            for channel in self.group_channels.values():
                users = channel.users
                for index in range(len(users)):
                    if users[index].id == client_id:
                        users[index] = alter_ego
                        continue
        
        self.private_channels.clear()
        self.group_channels.clear()
        self.events.clear()
        
        self.guild_profiles.clear()
        self.thread_profiles = None
        self.status = Status.offline
        self.statuses = None
        self._activity = ACTIVITY_UNKNOWN
        self.activities = None
        self.ready_state = None
    
    
    # login
    async def client_login_static(self):
        """
        The first step at login in is requesting the client's user data. This method is also used to check whether
        the token of the client is valid.
        
        This method is a coroutine.
        
        Returns
        -------
        response_data : `dict` of (`str` : `object`)
            Decoded json data got from Discord.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        InvalidToken
            When the token of the client is invalid.
        """
        while True:
            try:
                data = await self.api.client_user_get()
            except DiscordException as err:
                status = err.status
                if status == 401:
                    raise InvalidToken() from err
                
                if status >= 500:
                    await sleep(2.5, KOKORO)
                    continue
                
                raise
            
            return data
    
    
    async def update_application_info(self):
        """
        Updates the client's application's info.
        
        This method is a coroutine.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        Meanwhile the clients logs in this method is called to ensure that the client's application info is loaded.
        
        This endpoint is available only for bot accounts.
        """
        if not self.bot:
            return
        
        data = await self.api.application_get_own()
        application = self.application
        old_application_id = application.id
        application = application.from_data_own(data)
        self.application = application
        new_application_id = application.id
        
        if old_application_id != new_application_id:
            if APPLICATION_ID_TO_CLIENT.get(old_application_id, None) is self:
                del APPLICATION_ID_TO_CLIENT[old_application_id]
            
            APPLICATION_ID_TO_CLIENT[new_application_id] = self
        
        await self.emoji_get_all_application()
    
    
    async def client_gateway(self):
        """
        Requests the gateway information for the client.
        
        Only `1` request can be done at a time and every other will yield the result of first started one.
        
        This method is a coroutine.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        
        Raises
        ------
        ConnectionError
            No internet connection or if the request raised any ``DiscordException``.
        InvalidToken
            When the client's token is invalid.
        DiscordException
            If any exception was received from the Discord API.
        """
        if self._gateway_requesting:
            gateway_waiter = self._gateway_waiter
            if gateway_waiter is None:
                gateway_waiter = self._gateway_waiter = Future(KOKORO)
            
            return await gateway_waiter
        
        self._gateway_requesting = True
        
        try:
            while True:
                if self.bot:
                    coroutine = self.api.client_gateway_bot()
                else:
                    coroutine = self.api.client_gateway_hooman()
                try:
                    data = await coroutine
                except DiscordException as err:
                    status = err.status
                    if status == 401:
                        await self.disconnect()
                        raise InvalidToken() from err
                    
                    if status >= 500:
                        await sleep(2.5, KOKORO)
                        continue
                    
                    raise
                
                break
            
            # Note:
            # `max_concurrency` for small bots is 1000.
            # For large bots it is: `max(2000, (guild_count / 1000) * 5)`
            try:
                session_start_limit_data = data['session_start_limit']
            except KeyError:
                gateway_max_concurrency = 1
            else:
                gateway_max_concurrency = session_start_limit_data.get('max_concurrency', 1)
                
                try:
                    remaining = session_start_limit_data['remaining']
                except KeyError:
                    pass
                else:
                    if remaining < 100:
                        warnings.warn(
                            f'`Remaining session start limit reached low amount: {remaining!r}.',
                            RuntimeWarning,
                        )
            
            self._gateway_max_concurrency = gateway_max_concurrency
        except BaseException as err:
            self._gateway_requesting = False
            gateway_waiter = self._gateway_waiter
            if (gateway_waiter is not None):
                self._gateway_waiter = None
                gateway_waiter.set_exception_if_pending(err)
            
            raise
        
        self._gateway_requesting = False
        gateway_waiter = self._gateway_waiter
        if (gateway_waiter is not None):
            self._gateway_waiter = None
            gateway_waiter.set_result_if_pending(data)
        
        return data
    
    
    async def client_gateway_url(self):
        """
        Requests the client's gateway url. To avoid unreasoned requests when sharding, if this request was done at the
        last `60` seconds then returns the last generated url.
        
        This method is a coroutine.
        
        Raises
        ------
        ConnectionError
            No internet connection or if the request raised any ``DiscordException``.
        InvalidToken
            When the client's token is invalid.
        DiscordException
            If any exception was received from the Discord API.
        
        Returns
        -------
        gateway_url : `str` 
            The url to what the gateways' web socket will be connected.
        """
        if self._gateway_time > (LOOP_TIME() - 60.0):
            return self._gateway_url
        
        data = await self.client_gateway()
        self._gateway_url = gateway_url = data['url']
        self._gateway_time = LOOP_TIME()
        
        return gateway_url
    
    
    async def client_gateway_reshard(self, force = False):
        """
        Reshards the client. And also updates it's gateway's url as a side note.
        
        Should be called only if every shard is down.
        
        This method is a coroutine.
        
        Parameters
        ----------
        force : `bool`
            Whether the the client should reshard to lower amount of shards if needed.
        
        Raises
        ------
        ConnectionError
            No internet connection or if the request raised any ``DiscordException``.
        InvalidToken
            When the client's token is invalid.
        DiscordException
            If any exception was received from the Discord API.
        """
        data = await self.client_gateway()
        self._gateway_url = data['url']
        self._gateway_time = LOOP_TIME()
        
        old_shard_count = self.shard_count
        if old_shard_count <= 0:
            old_shard_count = 1
        
        new_shard_count = data.get('shards', 1)
        if new_shard_count <= 0:
            new_shard_count = 1
        
        # Do we have more shards already?
        if (not force) and (old_shard_count >= new_shard_count):
            return
        
        self.shard_count = new_shard_count
        self.gateway = reshard_gateway(self)
    
    
    def start(self):
        """
        Starts the clients' connecting to Discord. If the client is already running, raises `RuntimeError`.
        
        The return of the method depends on the thread, from which it was called from.
        
        Returns
        -------
        task : `bool`, ``Task``, ``FutureWrapperAsync``
            - If the method was called from the client's thread (KOKORO), then returns a ``Task``. The task will return
                `True`, if connecting was successful.
            - If the method was called from an ``EventThread``, but not from the client's, then returns a
                `FutureWrapperAsync`. The task will return `True`, if connecting was successful.
            - If the method was called from any other thread, then waits for the connector task to finish and returns
                `True`, if it was successful.
        
        Raises
        ------
        RuntimeError
            If the client is already running.
        """
        if self.running:
            raise RuntimeError(f'{self!r} is already running!')
        
        return run_coroutine(self.connect(), KOKORO)
    
    
    def stop(self):
        """
        Starts disconnecting the client.
        
        The return of the method depends on the thread, from which it was called from.
        
        Returns
        -------
        task : `None`, ``Task``, ``FutureWrapperAsync``
            - If the method was called from the client's thread (KOKORO), then returns a ``Task``.
            - If the method was called from an ``EventThread``, but not from the client's, then returns a
                `FutureWrapperAsync`.
            - If the method was called from any other thread, returns `None` when disconnecting finished.
        """
        return run_coroutine(self.disconnect(), KOKORO)
    
    
    async def connect(self):
        """
        Starts connecting the client to Discord, fills up the undefined events and creates the task, what will keep
        receiving the data from Discord (``._connect``).
        
        If you want to start the connecting process consider using the top-level ``.start``, ``start_clients`` instead.
        
        This method is a coroutine.
        
        Returns
        -------
        success : `bool`
            Whether the client could be started.
        
        Raises
        ------
        RuntimeError
            If the client is already running.
        """
        if self.running:
            raise RuntimeError(f'{self!r} is already running!')
        
        try:
            data = await self.client_login_static()
        except GeneratorExit:
            raise
        
        except BaseException as err:
            if isinstance(err, ConnectionError) and err.args[0] == 'Invalid address':
                after = (
                    'Connection failed, could not connect to Discord.\nPlease check your internet connection / has '
                    'Python rights to use it?\n'
                )
            else:
                after = None
            
            before = [
                'Exception occurred at calling ',
                type(self).__name__,
                '.connect\n',
            ]
            
            await write_exception_async(err, before, after, loop = KOKORO)
            return False
        
        # Some Discord implementations send string response for some weird reason.
        if isinstance(data, str):
            try:
                data = from_json(data)
            except JSONDecodeError:
                pass
        
        if not isinstance(data, dict):
            sys.stderr.write(
                ''.join([
                    (
                        'Connection failed, could not connect to Discord.\n'
                        'Received invalid data:\n'
                    ),
                    repr(data),
                    '\n',
                ])
            )
            return False
        
        self._init_on_ready(data)
        
        await self.client_gateway_reshard()
        
        if self.bot:
            task = Task(KOKORO, self.update_application_info())
            task.silence()
        
        # Check it twice, because meanwhile logging in, connect calls are not limited
        if self.running:
            raise RuntimeError(f'{self!r} is already running!')
        
        self.running = True
        register_client(self)
        Task(KOKORO, self._connect())
        return True
    
    
    async def _connect(self):
        """
        Connects the client's gateway(s) to Discord and reconnects them if needed.
        
        This method is a coroutine.
        """
    
        try:
            while True:
                ready_state = self.ready_state
                if (ready_state is not None):
                    self.ready_state = None
                    ready_state.cancel()
                    ready_state = None
                
                try:
                    await self.gateway.run()
                except CancelledError as err:
                    # For now only here. These errors occurred randomly for me since I made the wrapper, only once-once,
                    # and it was not the wrapper causing them, so it is time to say STOP.
                    # I also know `GeneratorExit` will show up as RuntimeError, but it is already a RuntimeError.
                    write_exception_async(
                        err,
                        [
                            'Ignoring unexpected outer Task or coroutine cancellation at ',
                            repr(self),
                            '._connect:\n',
                        ],
                        loop = KOKORO,
                    )
                    continue
                
                except DiscordGatewayException as err:
                    if err.code in RESHARD_ERROR_CODES:
                        sys.stderr.write(
                            f'{err.__class__.__name__} occurred, at {self!r}._connect:\n'
                            f'{err!r}\n'
                            f'The client will reshard itself and reconnect.\n'
                        )
                        
                        await self.client_gateway_reshard(force = True)
                        continue
                    
                    raise
                
                else:
                    if not self.running:
                        break
                    
                    while True:
                        try:
                            await sleep(5.0, KOKORO)
                            try:
                                # We are down, why not reshard instantly?
                                await self.client_gateway_reshard()
                            except ConnectionError:
                                continue
                            else:
                                break
                        except CancelledError as err:
                            write_exception_async(
                                err,
                                [
                                    'Ignoring unexpected outer Task or coroutine cancellation at ',
                                    repr(self),
                                    '._connect:\n',
                                ],
                                loop = KOKORO,
                            )
                            continue
                    
                    continue
        except GeneratorExit:
            raise
        except BaseException as err:
            if (
                isinstance(err, InvalidToken) or
                (
                    isinstance(err, DiscordGatewayException) and
                    (err.code in INTENT_ERROR_CODES)
                )
            ):
                sys.stderr.write(
                    f'{err.__class__.__name__} occurred, at {self!r}._connect:\n'
                    f'{err!r}\n'
                )
            else:
                write_exception_async(
                    err,
                    [
                        'Unexpected exception occurred at ',
                        repr(self),
                        '._connect\n',
                    ],
                    (
                        'If you can reproduce this bug, Please send me a message or open an issue with your code, and '
                        'with every detail how to reproduce it.\n'
                        'Thanks!\n'
                    ),
                    loop = KOKORO,
                )
            
            await ensure_shutdown_event_handlers(self)
        
        finally:
            try:
                self.gateway.abort()
            finally:
                unregister_client(self)
                self.running = False
                
                if not self.guild_profiles:
                    return
                
                for guild_id in [*self.guild_profiles.keys()]:
                    try:
                        guild = GUILDS[guild_id]
                    except KeyError:
                        continue
                    
                    guild._delete(self)
                
                # need to delete the references for cleanup
                guild = None
                
                ready_state = self.ready_state
                if (ready_state is not None):
                    self.ready_state = None
                    ready_state.cancel()
                    ready_state = None
    
    
    def _delay_ready(self, guild_datas, shard_id):
        """
        Delays the client's "ready" till it receives all of it guild's data. If caching is allowed (so by default),
        then it waits additional time till it requests all the members of it's guilds.
        
        Parameters
        ----------
        guild_datas : `list` of `dict` (`str`, `object`) items
            Partial data for all the guilds to request the users of.
        shard_id : `int`
            The received shard's identifier.
        """
        ready_state = self.ready_state
        if (ready_state is None):
            ready_state = ReadyState(self)
            self.ready_state = ready_state
        
        ready_state.shard_ready(self, guild_datas, shard_id)
    
    
    
    async def disconnect(self):
        """
        Disconnects the client and closes it's web socket(s). Till the client goes offline, it might take even over
        than `1` minute. Because bot accounts can not logout, so they need to wait for timeout.
        
        This method is a coroutine.
        """
        if not self.running:
            return
        
        self.running = False
        
        await ensure_voice_client_shutdown_event_handlers(self)
        
        # Log off if user account
        if (not self.bot):
            await self.api.client_logout()
        
        # Close gateways
        # cancel shards
        await self.gateway.close()
        
        await ensure_shutdown_event_handlers(self)
    
    
    def voice_client_for(self, message):
        """
        Returns the voice client for the given message's guild if it has any.
        
        Parameters
        ----------
        message : ``Message``
            The message what's voice client will be looked up.
        
        Returns
        -------
        voice_client : `None`, ``VoiceClient``
            The voice client if applicable.
        """
        guild_id = message.guild_id
        if guild_id is None:
            voice_client = None
        else:
            voice_client = self.voice_clients.get(guild_id, None)
        return voice_client
    
    
    def get_guild(self, name, default = None):
        """
        Tries to find a guild by it's name. If there is no guild with the given name, then returns the passed
        default value.
        
        Parameters
        ----------
        name : `str`
            The guild's name to search.
        default : `None`, `object` = `None`, Optional
            The default value, what will be returned if the guild was not found.
        
        Returns
        -------
        guild : ``Guild``, `default`
        
        Raises
        ------
        AssertionError
            - If `name` was not given as `str`.
            - If `name` length is out of the expected range [2:100].
        
        """
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                )
            
            name_length = len(name)
            if name_length < 2 or name_length > 100:
                raise AssertionError(
                    f'`name` length can be in range [1:100], got {name_length!r}; {name!r}.'
                )
        
        for guild_id in self.guild_profiles.keys():
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                continue
            
            if guild.name == name:
                return guild
        
        return default
    
    
    get_rate_limits_of = methodize(RateLimitProxy)
    
    
    @property
    def owner(self):
        """
        Returns the client's owner if applicable.
        
        If the client is a user account, or if it's ``.update_application_info`` was not called yet, then returns
        ``ZEROUSER``. If the client is owned by a ``Team``, then returns the team's owner.
        
        Returns
        -------
        owner : ``ClientUserBase``
        """
        application_owner = self.application.owner
        if isinstance(application_owner, Team):
            owner = application_owner.owner
        else:
            owner = application_owner
        
        return owner
    
    
    def is_owner(self, user):
        """
        Returns whether the passed user is one of the client's owners.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The user who will be checked.
        
        Returns
        -------
        is_owner : `bool`
        """
        application_owner = self.application.owner
        if isinstance(application_owner, Team):
            if user in application_owner.accepted:
                return True
        else:
            if application_owner is user:
                return True
        
        additional_owner_ids = self._additional_owner_ids
        if (additional_owner_ids is not None) and (user.id in additional_owner_ids):
            return True
        
        return False
    
    
    def add_additional_owners(self, *users):
        """
        Adds additional users to be passed at the ``.is_owner`` check.
        
        Parameters
        ----------
        *users : `int`, ``UserBase``
            The `.id` of the a user or the user itself to be added.
        
        Raises
        ------
        TypeError
            A user was passed with invalid type.
        """
        limit = len(users)
        if limit == 0:
            return
        
        index = 0
        while True:
            user = users[index]
            index += 1
            if not isinstance(user, (int, UserBase)):
                raise TypeError(
                    f'`users[{index}]` is not `int`, `{UserBase.__name__}`, got {user.__class__.__name__}; {user!r}; '
                    f'users = {users!r}.'
                )
            
            if index == limit:
                break
        
        additional_owner_ids = self._additional_owner_ids
        
        for user in users:
            if type(user) is int:
                user_id = user
            
            elif isinstance(user, int):
                user_id = int(user)
            
            else:
                user_id = user.id
            
            if additional_owner_ids is None:
                additional_owner_ids = set()
                self._additional_owner_ids = additional_owner_ids
            
            additional_owner_ids.add(user_id)
    
    
    def remove_additional_owners(self, *users):
        """
        Removes additional owners added by the ``.add_additional_owners`` method.
        
        Parameters
        ----------
        *users : `int`, ``UserBase``
            The `.id` of the a user or the user itself to be removed.
        
        Raises
        ------
        TypeError
            A user was passed with invalid type.
        """
        limit = len(users)
        if limit == 0:
            return
        
        index = 0
        while True:
            user = users[index]
            index += 1
            if not isinstance(user, (int, UserBase)):
                raise TypeError(
                    f'`users[{index}]` is not `int`, `{UserBase.__name__}`, got {user.__class__.__name__}; {user!r}; '
                    f'users = {users!r}.'
                )
            
            if index == limit:
                break
        
        additional_owner_ids = self._additional_owner_ids
        
        for user in users:
            if type(user) is int:
                user_id = user
            
            elif isinstance(user, int):
                user_id = int(user)
            
            else:
                user_id = user.id
            
            if additional_owner_ids is None:
                continue
            
            additional_owner_ids.discard(user_id)
            if not additional_owner_ids:
                additional_owner_ids = None
                self._additional_owner_ids = None
    
    
    @property
    def owners(self):
        """
        Returns the owners of the client.
        
        Returns
        -------
        owners : `set` of ``ClientUserBase``
        """
        owners = set()
        
        application_owner = self.application.owner
        if type(application_owner) is Team:
            owners.update(application_owner.accepted)
        else:
            owners.add(application_owner)
        
        additional_owner_ids = self._additional_owner_ids
        if (additional_owner_ids is not None):
            for user_id in additional_owner_ids:
                user = create_partial_user_from_id(user_id)
                owners.add(user)
        
        return owners
    
    
    @copy_docs(ClientUserPBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ClientUserPBase._difference_update_attributes(self, data)
        
        email = parse_email(data)
        if self.email != email:
            old_attributes['email'] = self.email
            self.email = email
        
        email_verified = parse_email_verified(data)
        if self.email_verified != email_verified:
            old_attributes['email_verified'] = self.email_verified
            self.email_verified = email_verified
        
        locale = parse_locale(data)
        if self.locale is not locale:
            old_attributes['locale'] = self.locale
            self.locale = locale
        
        mfa_enabled = parse_mfa_enabled(data)
        if self.mfa_enabled != mfa_enabled:
            old_attributes['mfa_enabled'] = self.mfa_enabled
            self.mfa_enabled = mfa_enabled
        
        premium_type = parse_premium_type(data)
        if self.premium_type is not premium_type:
            old_attributes['premium_type'] = self.premium_type
            self.premium_type = premium_type
        
        return old_attributes
    
    
    @copy_docs(ClientUserPBase._update_attributes)
    def _update_attributes(self, data):
        ClientUserPBase._update_attributes(self, data)
        
        self.email = parse_email(data)
        self.email_verified = parse_email_verified(data)
        self.locale = parse_locale(data)
        self.mfa_enabled = parse_mfa_enabled(data)
        self.premium_type = parse_premium_type(data)
    
    
    @copy_docs(ClientUserPBase._difference_update_profile)
    def _difference_update_profile(self, data, guild):
        old_attributes = ClientUserPBase._difference_update_profile(self, data, guild)
        guild._invalidate_cache_permission()
        return old_attributes
    
    
    @copy_docs(ClientUserPBase._update_profile)
    def _update_profile(self, data, guild):
        updated_or_created = ClientUserPBase._update_profile(self, data, guild)
        guild._invalidate_cache_permission()
        return updated_or_created
    
    
    @property
    def friends(self):
        """
        Returns the client's friends.
        
        Returns
        -------
        relationships : `list` of ``Relationship`` objects
        """
        type_ = RelationshipType.friend
        return [relationship for relationship in self.relationships.values() if relationship.type is type_]
    
    
    @property
    def blocked(self):
        """
        Returns the client's blocked relationships.
        
        Returns
        -------
        relationships : `list` of ``Relationship`` objects
        """
        type_ = RelationshipType.blocked
        return [relationship for relationship in self.relationships.values() if relationship.type is type_]
    
    
    @property
    def received_requests(self):
        """
        Returns the received friend requests of the client.
        
        Returns
        -------
        relationships : `list` of ``Relationship`` objects
        """
        type_ = RelationshipType.pending_incoming
        return [relationship for relationship in self.relationships.values() if relationship.type is type_]
    
    
    @property
    def sent_requests(self):
        """
        Returns the sent friend requests of the client.
        
        Returns
        -------
        relationships : `list` of ``Relationship`` objects
        """
        type_ = RelationshipType.pending_outgoing
        return [relationship for relationship in self.relationships.values() if relationship.type is type_]
    
    
    def gateway_for(self, guild_id):
        """
        Returns the corresponding gateway of the client to the passed guild.
        
        Parameters
        ----------
        guild_id : `int`
            The respective guild's identifier, what's gateway will be returned.
            
            Pass it as `0` to get the default gateway.
        
        Returns
        -------
        gateway : ``DiscordGatewayClientBase``
        """
        return self.gateway.get_gateway(guild_id)
    
    
    @classmethod
    @copy_docs(ClientUserBase._from_client)
    def _from_client(cls, client, include_internals):
        raise RuntimeError('Cannot create client copy from client.')
    
    
    @classmethod
    @copy_docs(ClientUserBase._create_empty)
    def _create_empty(cls, user_id):
        raise RuntimeError('Cannot create empty client.')
    
    
    @copy_docs(ClientUserBase.copy)
    def copy(self):
        return User._from_client(self, False)
    
    
    @copy_docs(ClientUserBase.copy_with)
    def copy_with(self, **keyword_parameters):
        return User._from_client(self, False).copy_with(**keyword_parameters)



__all__ = ('CommandBase',)

from scarletio import RichAttributeErrorBaseType

from ...interfaces.exception_handler import ExceptionHandlerInterface


class CommandBase(ExceptionHandlerInterface, RichAttributeErrorBaseType):
    """
    Base type for ``Slasher``'s commands.
    
    Attributes
    ----------
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
        
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>`
        The parent slasher of the component command.

    name : `str`
        The command's name.
        
        Only used for debugging.
    """
    __slots__ = ('_exception_handlers', '_parent_reference', 'name')
    
    
    def __new__(cls, function, name = None, **keyword_parameters):
        """
        Creates a new command instance.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        function : `None | async-callable`
            The function used as the command when using the respective slash command.
        
        name : `None | str` = `None`, Optional
            The name of the component command.
        
        **keyword_parameters : Keyword Parameters
            Additional Keyword parameters.
        
        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError
    
    
    def __repr__(self):
        """Returns the command's representation."""
        repr_parts = ['<', type(self).__name__]
        repr_parts = self._put_repr_parts_into(repr_parts)
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _put_repr_parts_into(self, repr_parts):
        """
        Representation builder helper to build the representation's body.
        
        Parameters
        ----------
        repr_parts : `list<str>`
            Parts to extend.
        
        Returns
        -------
        repr_parts : `list<str>`
        """
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            repr_parts.append(', exception_handlers = ')
            repr_parts.append(repr(exception_handlers))
        
        return repr_parts
    
    
    def __hash__(self):
        """Returns the command's hash value."""
        hash_value = 0
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            hash_value ^= len(exception_handlers) << 4
            
            for exception_handler in exception_handlers:
                try:
                    exception_handler_hash_value = hash(exception_handler)
                except TypeError:
                    exception_handler_hash_value = object.__hash__(exception_handler)
                hash_value ^= exception_handler_hash_value
        
        # _parent_reference
        # Internal field
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two commands are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # _exception_handlers
        if self._exception_handlers != other._exception_handlers:
            return False
        
        # _parent_reference
        # Internal field
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __format__(self, code):
        """Formats the command in a format string."""
        if not code:
            return str(self)
        
        if code == 'm':
            return self.mention
        
        raise ValueError(
            f'Unknown format code {code!r} for {type(self).__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"m"!r}.'
        )
    
    
    async def invoke(self, client, interaction_event):
        """
        Calls the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        
        interaction_event : ``InteractionEvent``
            The received interaction event.
        """
        return
    
    
    def copy(self):
        """
        Copies the command.
        
        Returns
        -------
        new : `type<self>`
        """
        new = object.__new__(type(self))
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        # _parent_reference
        new._parent_reference = None
        
        # name
        new.name = self.name
        
        return new
    
    
    # ---- Mention ----
    
    @property
    def mention(self):
        """
        Returns the command mention.
        Applicable for slash and context commands.
        
        Returns
        -------
        mention : `str`
        """
        return ''
    
    
    def mention_at(self, guild):
        """
        Mentions the command at the specified guild.
        Should be used when the command is added to multiple guilds.
        Applicable for slash and context commands.
        
        Parameters
        ----------
        guild : `Guild | int`
            The guild to get the command's mention.
        
        Returns
        -------
        mention : `str`
        """
        return ''



__all__ = ('RoleManagerType',)

from ...bases import Preinstance as P, PreinstancedBase

from ..role_manager_metadata import (
    RoleManagerMetadataApplicationRoleConnection, RoleManagerMetadataBase, RoleManagerMetadataBooster,
    RoleManagerMetadataBot, RoleManagerMetadataIntegration, RoleManagerMetadataSubscription
)

class RoleManagerType(PreinstancedBase, value_type = int):
    """
    Represents a managed role's manager type.
    
    Attributes
    ----------
    name : `str`
        The name of the role manager type.
    
    value : `int`
        The identifier value the role manager type.
    
    Type Attributes
    ---------------
    Every predefined role manager type can be accessed as type attribute as well:
    
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | Type attribute name           | Name                          | Value | Metadata type                                     |
    +===============================+===============================+=======+===================================================+
    | none                          | none                          | 0     | ``RoleManagerMetadataBase``                       |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | unset                         | unset                         | 1     | ``RoleManagerMetadataBase``                       |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | unknown                       | unknown                       | 2     | ``RoleManagerMetadataBase``                       |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | bot                           | bot                           | 3     | ``RoleManagerMetadataBot``                        |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | booster                       | booster                       | 4     | ``RoleManagerMetadataBooster``                    |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | integration                   | integration                   | 5     | ``RoleManagerMetadataIntegration``                |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | subscription                  | subscription                  | 6     | ``RoleManagerMetadataSubscription``               |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    | application_role_connection   | application role connection   | 7     | ``RoleManagerMetadataApplicationRoleConnection``  |
    +-------------------------------+-------------------------------+-------+---------------------------------------------------+
    """
    __slots__ = ('metadata_type', )
    
    
    def __new__(cls, value, name = None, metadata_type = None):
        """
        Creates a new scheduled event entity type instance from the given parameters.
        
        Parameters
        ----------
        value : `str`
            The unique identifier of the scheduled event entity type.
        
        name : `None | str` = `None`, Optional
            The name of the scheduled event entity type.
        
        metadata_type : `None | type<RoleManagerMetadataBase>` = `None`, Optional
            The role manager's metadata's type.
        """
        if metadata_type is None:
            metadata_type = RoleManagerMetadataBase
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.metadata_type = metadata_type
        return self
    
    
    def __bool__(self):
        """Returns whether the role manager's type is set."""
        if self.value:
            boolean = True
        else:
            boolean = False
        
        return boolean
    
    
    none = P(0, 'none', RoleManagerMetadataBase)
    unset = P(1, 'unset', RoleManagerMetadataBase)
    unknown = P(2, 'unknown', RoleManagerMetadataBase)
    bot = P(3, 'bot', RoleManagerMetadataBot)
    booster = P(4, 'booster', RoleManagerMetadataBooster)
    integration = P(5, 'integration', RoleManagerMetadataIntegration)
    subscription = P(6, 'subscription', RoleManagerMetadataSubscription)
    application_role_connection = P(7, 'application role connection', RoleManagerMetadataApplicationRoleConnection)



__all__ = ('Webhook', )

from scarletio import copy_docs, export, include

from ...bases import ICON_TYPE_NONE, Icon
from ...core import USERS
from ...exceptions import DiscordException, ERROR_CODES
from ...http import urls as module_urls
from ...http.urls import WEBHOOK_URL_PATTERN
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ClientUserBase, UserBase, ZEROUSER
from ...user.user.fields import parse_id, put_id, put_webhook_name, validate_id, validate_webhook_name

from ..webhook_source_channel import WebhookSourceChannel
from ..webhook_source_guild import WebhookSourceGuild

from .fields import (
    parse_application_id, parse_channel_id, parse_source_channel, parse_source_guild, parse_token, parse_type,
    parse_user, put_application_id, put_channel_id, put_source_channel, put_source_guild,
    put_token, put_type, put_user, validate_application_id, validate_channel_id,
    validate_source_channel, validate_source_guild, validate_token, validate_type, validate_user
)
from .preinstanced import WebhookType
from .webhook_base import WebhookBase


create_partial_webhook_from_id = include('create_partial_webhook_from_id')


PRECREATE_FIELDS = {
    'application': ('application_id', validate_application_id),
    'application_id': ('application_id', validate_application_id),
    'avatar': ('avatar', UserBase.avatar.validate_icon),
    'channel': ('channel_id', validate_channel_id),
    'channel_id': ('channel_id', validate_channel_id),
    'name': ('name', validate_webhook_name),
    'source_channel': ('source_channel', validate_source_channel),
    'source_guild': ('source_guild', validate_source_guild),
    'token': ('token', validate_token),
    'user': ('user', validate_user),
    'webhook_type': ('type', validate_type),
}


@export
class Webhook(WebhookBase):
    """
    Represents a Discord webhook. At some cases it might be used as webhook's user representation.
    
    Attributes
    ----------
    application_id : `int`
        The application's identifier what created the webhook. Defaults to `0` if not applicable.
    avatar_hash : `int`
        The webhook's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The webhook's avatar's type.
    channel_id : `int`
        The channel's identifier, where the webhook is going to send it's messages.
    id : `int`
        The webhook's unique identifier number.
    name : str
        The webhook's name.
    source_channel : `None`, ``WebhookSourceChannel``
        Representation of the webhook's source channel. Applicable for ``WebhookType.server`` webhooks.
    source_guild : `None`, ``WebhookSourceGuild``
        Representation of the webhook's source guild. Applicable for ``WebhookType.server`` webhooks.
    token : `str`
        The webhook's token. You need an `id` and a `token` to send webhook message. Defaults to empty string.
    type : ``WebhookType``
        The webhook's type.
    user : ``ClientUserBase``
        The creator of the webhook, or `ZEROUSER` if unknown.
    
    Notes
    -----
    Instances of this class are weakreferable.
    """
    __slots__ = ('application_id', 'source_channel', 'source_guild', 'token', 'user')
    
    def __new__(
        cls,
        *,
        application_id = ...,
        avatar = ...,
        channel_id = ...,
        source_channel = ...,
        source_guild = ...,
        name = ...,
        token = ...,
        user = ...,
        webhook_type = ...,
    ):
        """
        Creates a new partial webhook base with the given fields.
        
        Parameters
        ----------
        application_id : `int`, ``Application``, Optional (Keyword only)
            The application's identifier what created the webhook.
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier, where the webhook is going to send it's messages.
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        source_channel : `None`, ``WebhookSourceChannel``, Optional (Keyword only)
            Representation of the webhook's source channel.
        source_guild : `None`, ``WebhookSourceGuild``, Optional (Keyword only)
            Representation of the webhook's source guild.
        name : `str`, Optional (Keyword only)
            The user's name.
        token : `str`, Optional (Keyword only)
            The webhook's token.
        user : ``ClientUserBase``, Optional (Keyword only)
            The creator of the webhook.
        webhook_type : ``WebhookType``, `int`, Optional (Keyword only)
            The webhook's type.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_id
        if application_id is ...:
            application_id = 0
        else:
            application_id = validate_application_id(application_id)
        
        # source_channel
        if source_channel is ...:
            source_channel = None
        else:
            source_channel = validate_source_channel(source_channel)
        
        # source_guild
        if source_guild is ...:
            source_guild = None
        else:
            source_guild = validate_source_guild(source_guild)
        
        # token
        if token is ...:
            token = ''
        else:
            token = validate_token(token)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # Construct
        self = WebhookBase.__new__(
            cls,
            avatar = avatar,
            channel_id = channel_id,
            name = name,
            webhook_type = webhook_type,
        )
        self.application_id = application_id
        self.source_channel = source_channel
        self.source_guild = source_guild
        self.token = token
        self.user = user
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Tries to get the webhook from the existing ones, then update it. If no webhook was found, creates a new one and
        fills it's attributes from the data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received webhook data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        webhook_id = parse_id(data)
        try:
            self = USERS[webhook_id]
        except KeyError:
            self = object.__new__(cls)
            USERS[webhook_id] = self
            self.id = webhook_id
            self.token = ''
        
        self._set_attributes(data)
        
        return self
    
    
    def to_webhook_data(self, *, defaults = False, include_internals = False):
        """
        Tries to convert the webhook back to a json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        type(self).avatar.put_into(self.avatar, data, defaults, as_data = not include_internals)
        put_channel_id(self.channel_id, data, defaults)
        put_webhook_name(self.name, data, defaults)
        
        if include_internals:
            put_application_id(self.application_id, data, defaults)
            put_id(self.id, data, defaults)
            put_source_channel(self.source_channel, data, defaults)
            put_source_guild(self.source_guild, data, defaults)
            put_token(self.token, data, defaults)
            put_type(self.type, data, defaults)
            put_user(self.user, data, defaults)
        
        return data
    
    
    @copy_docs(UserBase.copy)
    def copy(self):
        new = WebhookBase.copy(self)
        new.application_id = self.application_id
        new.channel_id = self.channel_id
        source_channel = self.source_channel
        if (source_channel is not None):
            source_channel = source_channel.copy()
        new.source_channel = source_channel
        source_guild = self.source_guild
        if (source_guild is not None):
            source_guild = source_guild.copy()
        new.source_guild = source_guild
        new.token = self.token
        new.user = self.user
        return new
    
    
    def copy_with(
        self,
        *,
        application_id = ...,
        avatar = ...,
        channel_id = ...,
        source_channel = ...,
        source_guild = ...,
        name = ...,
        token = ...,
        user = ...,
        webhook_type = ...,
    ):
        """
        Copies the webhook base with the given fields.
        
        Parameters
        ----------
        application_id : `int`, ``Application``, Optional (Keyword only)
            The application's identifier what created the webhook.
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier, where the webhook is going to send it's messages.
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        source_channel : `None`, ``WebhookSourceChannel``, Optional (Keyword only)
            Representation of the webhook's source channel.
        source_guild : `None`, ``WebhookSourceGuild``, Optional (Keyword only)
            Representation of the webhook's source guild.
        name : `str`, Optional (Keyword only)
            The user's name.
        token : `str`, Optional (Keyword only)
            The webhook's token.
        user : ``ClientUserBase``, Optional (Keyword only)
            The creator of the webhook.
        webhook_type : ``WebhookType``, `int`, Optional (Keyword only)
            The webhook's type.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_id
        if application_id is ...:
            application_id = self.application_id
        else:
            application_id = validate_application_id(application_id)
        
        # source_channel
        if source_channel is ...:
            source_channel = self.source_channel
            if (source_channel is not None):
                source_channel = source_channel.copy()
        else:
            source_channel = validate_source_channel(source_channel)
        
        # source_guild
        if source_guild is ...:
            source_guild = self.source_guild
            if (source_guild is not None):
                source_guild = source_guild.copy()
        else:
            source_guild = validate_source_guild(source_guild)
        
        # token
        if token is ...:
            token = self.token
        else:
            token = validate_token(token)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # Construct
        new = WebhookBase.copy_with(
            self,
            avatar = avatar,
            channel_id = channel_id,
            name = name,
            webhook_type = webhook_type,
        )
        
        new.application_id = application_id
        new.source_channel = source_channel
        new.source_guild = source_guild
        new.token = token
        new.user = user
        return new
    
    
    @copy_docs(WebhookBase._get_hash_partial)
    def _get_hash_partial(self):
        hash_value = WebhookBase._get_hash_partial(self)
        
        # application_id
        hash_value ^= self.application_id
        
        # source_channel
        source_channel = self.source_channel
        if (source_channel is not None):
            hash_value ^= hash(source_channel)
        
        # source_guild
        source_guild = self.source_guild
        if (source_guild is not None):
            hash_value ^= hash(source_guild)
        
        # token
        hash_value ^= hash(self.token)
        
        # user
        user = self.user
        if (user is not ZEROUSER):
            hash_value ^= hash(user)
        
        return user
    
    
    def _set_attributes(self, data):
        """
        Sets the webhook's attributes from the given data.
        
        Not like ``._update_attributes``, this method also sets the attributes that are not expected to be modified.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received webhook data.
        """
        self._update_attributes(data)
        
        self.application_id = parse_application_id(data)
        self.source_channel = parse_source_channel(data)
        self.source_guild = parse_source_guild(data)
        
        token = parse_token(data)
        if token:
            self.token = token
        
        self.type = parse_type(data)
        self.user = parse_user(data)
    
    
    @copy_docs(WebhookBase._update_attributes)
    def _update_attributes(self, data):
        WebhookBase._update_attributes(self, data)
        self.channel_id = parse_channel_id(data)
    
    
    @copy_docs(WebhookBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = WebhookBase._difference_update_attributes(self, data)
        
        channel_id = parse_channel_id(data)
        if channel_id != self.channel_id:
            old_attributes['channel_id'] = self.channel_id
            self.channel_id = channel_id
        
        return old_attributes
    
    
    @copy_docs(UserBase._set_default_attributes)
    def _set_default_attributes(self):
        WebhookBase._set_default_attributes(self)
        
        self.application_id = 0
        self.source_guild = None
        self.source_channel = None
        self.token = ''
        self.user = ZEROUSER
    
    
    @classmethod
    def precreate(cls, webhook_id, **keyword_parameters):
        """
        Precreates the webhook by creating a partial one with the given parameters. When the webhook will be loaded,
        the precreated one will be picked up and will be updated. If an already existing webhook would be precreated,
        it will be updated with the given parameters only if it is a partial one.
        
        Parameters
        ----------
        webhook_id : `int`, `str`
            The webhook's id.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the webhook.
        
        Other Parameters
        ----------------
        application : `int`, ``Application``, Optional (Keyword only)
            Alternative for `application_id`.
        
        application_id : `int`, ``Application``, Optional (Keyword only)
            The application's identifier what created the webhook.
        
        avatar : `None`, ``Icon``, `str`, Optional (Keyword only)
            The webhook's avatar.
        
        channel : `int`, ``Channel``, Optional (Keyword only)
            Alternative for `channel_id`.
        
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel's identifier, where the webhook is going to send it's messages.
        
        name : `str`, Optional (Keyword only)
            The webhook's name.
        
        source_channel : `None`, ``WebhookSourceChannel``, Optional (Keyword only)
            Representation of the webhook's source channel.
        
        source_guild : `None`, ``WebhookSourceGuild``, Optional (Keyword only)
            Representation of the webhook's source guild.
        
        token : `str`, Optional (Keyword only)
            The webhook's token.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The creator of the webhook.
        
        webhook_type : ``WebhookType``, `int`, Optional (Keyword only)
            The webhook's type.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        webhook_id = validate_id(webhook_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = USERS[webhook_id]
        except KeyError:
            self = cls._create_empty(webhook_id)
            USERS[webhook_id] = self
        else:
            if not self.partial:
                return self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def from_url(cls, url):
        """
        Tries to parse the webhook's `id` and `token` from the given `url`. If succeeds, returns a partial webhook.
        If parsing fails, returns `None`.
        
        Parameters
        ----------
        url : `str`
            The url of the webhook.
        
        Returns
        -------
        self : `None`, `instance<cls>`
        """
        result = WEBHOOK_URL_PATTERN.fullmatch(url)
        if result is None:
            return None
        
        webhook_id = int(result.group(1))
        webhook_token = result.group(2)
        
        return create_partial_webhook_from_id(webhook_id, webhook_token)
    
    
    url = property(module_urls.webhook_url)
    
    
    @classmethod
    async def _from_follow_data(cls, data, source_channel, target_channel_id, client):
        """
        Creates the webhook, what executes cross-posts.
        
        This method is ensured after following a channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received webhook data.
        source_channel : ``Channel``
            The followed channel.
        target_channel_id : `int`
            The target channel's identifier where the webhook messages will be sent.
        client : ``Client``
            The client who created the webhook.
        
        Returns
        -------
        self : `instance<cls>`
        """
        webhook_id = int(data['webhook_id'])
        
        guild = source_channel.guild
        if guild is None:
            try:
                extra_data = await client.api.webhook_get(webhook_id)
            except DiscordException as err:
                if err.code == ERROR_CODES.unknown_webhook:
                    # not lucky
                    name = ''
                    avatar_type = ICON_TYPE_NONE
                    avatar_hash = 0
                else:
                    raise
            
            else:
                name = extra_data['name']
                if name is None:
                    name = ''
                
                avatar_type, avatar_hash = Icon.from_base_16_hash(data.get('avatar', None))
            
            source_guild = parse_source_guild(data)
        
        else:
            avatar_hash = guild.icon_hash
            avatar_type = guild.icon_type
            name = f'{guild.name} #{source_channel.name}'
            
            source_guild = WebhookSourceGuild.from_guild(guild)
        
        source_channel = WebhookSourceChannel.from_channel(source_channel)
        
        self = object.__new__(cls)
        self.application_id = 0
        self.avatar_hash = avatar_hash
        self.avatar_type = avatar_type
        self.channel_id = target_channel_id
        self.id = webhook_id
        self.name = name
        self.source_channel = source_channel
        self.source_guild = source_guild
        self.token = ''
        self.type = WebhookType.server
        self.user = client
        
        USERS[webhook_id] = self
        
        return self



from .achievement import *
from .application import *
from .application_command import *
from .application_role_connection import *
from .auto_moderation import *
from .channel import *
from .client import *
from .client_gateway import *
from .discovery import *
from .emoji_application import *
from .emoji_guild import *
from .guild import *
from .guild_ban import *
from .interaction import *
from .integration import *
from .invite import *
from .locked import *
from .message import *
from .oauth2 import *
from .miscellaneous import *
from .poll import *
from .reaction import *
from .role import *
from .scheduled_event import *
from .soundboard import *
from .stage import *
from .sticker import *
from .thread import *
from .user import *
from .webhook import *


__all__ = (
    *achievement.__all__,
    *application.__all__,
    *application_command.__all__,
    *application_role_connection.__all__,
    *auto_moderation.__all__,
    *channel.__all__,
    *client.__all__,
    *client_gateway.__all__,
    *discovery.__all__,
    *emoji_application.__all__,
    *emoji_guild.__all__,
    *guild.__all__,
    *guild_ban.__all__,
    *interaction.__all__,
    *integration.__all__,
    *invite.__all__,
    *locked.__all__,
    *message.__all__,
    *miscellaneous.__all__,
    *oauth2.__all__,
    *poll.__all__,
    *reaction.__all__,
    *role.__all__,
    *scheduled_event.__all__,
    *soundboard.__all__,
    *stage.__all__,
    *sticker.__all__,
    *thread.__all__,
    *user.__all__,
    *webhook.__all__,
)


from .achievement import ClientCompoundAchievementEndpoints
from .application import ClientCompoundApplicationEndpoints
from .application_command import ClientCompoundApplicationCommandEndpoints
from .application_role_connection import ClientCompoundApplicationRoleConnectionEndpoints
from .auto_moderation import ClientCompoundAutoModerationEndpoints
from .channel import ClientCompoundChannelEndpoints
from .client import ClientCompoundClientEndpoints
from .client_gateway import ClientCompoundClientGateway
from .discovery import ClientCompoundDiscoveryEndpoints
from .emoji_application import ClientCompoundEmojiApplicationEndpoints
from .emoji_guild import ClientCompoundEmojiGuildEndpoints
from .guild import ClientCompoundGuildEndpoints
from .guild_ban import ClientCompoundGuildBanEndpoints
from .interaction import ClientCompoundInteractionEndpoints
from .integration import ClientCompoundIntegrationEndpoints
from .invite import ClientCompoundInviteEndpoints
from .locked import ClientCompoundLockedEndpoints
from .message import ClientCompoundMessageEndpoints
from .miscellaneous import ClientCompoundMiscellaneousEndpoints
from .oauth2 import ClientCompoundOauth2Endpoints
from .poll import ClientCompoundPollEndpoints
from .reaction import ClientCompoundReactionEndpoints
from .role import ClientCompoundRoleEndpoints
from .scheduled_event import ClientCompoundScheduledEventEndpoints
from .soundboard import ClientCompoundSoundBoardEndpoints
from .stage import ClientCompoundStageEndpoints
from .sticker import ClientCompoundStickerEndpoints
from .thread import ClientCompoundThreadEndpoints
from .user import ClientCompoundUserEndpoints
from .webhook import ClientCompoundWebhookEndpoints


CLIENT_COMPOUNDS = (
    ClientCompoundAchievementEndpoints,
    ClientCompoundApplicationEndpoints,
    ClientCompoundApplicationCommandEndpoints,
    ClientCompoundApplicationRoleConnectionEndpoints,
    ClientCompoundAutoModerationEndpoints,
    ClientCompoundChannelEndpoints,
    ClientCompoundClientEndpoints,
    ClientCompoundClientGateway,
    ClientCompoundDiscoveryEndpoints,
    ClientCompoundEmojiApplicationEndpoints,
    ClientCompoundEmojiGuildEndpoints,
    ClientCompoundGuildEndpoints,
    ClientCompoundGuildBanEndpoints,
    ClientCompoundInteractionEndpoints,
    ClientCompoundIntegrationEndpoints,
    ClientCompoundInviteEndpoints,
    ClientCompoundLockedEndpoints,
    ClientCompoundMessageEndpoints,
    ClientCompoundMiscellaneousEndpoints,
    ClientCompoundOauth2Endpoints,
    ClientCompoundPollEndpoints,
    ClientCompoundReactionEndpoints,
    ClientCompoundRoleEndpoints,
    ClientCompoundScheduledEventEndpoints,
    ClientCompoundSoundBoardEndpoints,
    ClientCompoundStageEndpoints,
    ClientCompoundStickerEndpoints,
    ClientCompoundThreadEndpoints,
    ClientCompoundUserEndpoints,
    ClientCompoundWebhookEndpoints,
)



__all__ = ()

from ...channel import Channel
from ...field_parsers import (
    bool_parser_factory, entity_id_array_parser_factory, entity_id_parser_factory, force_string_parser_factory,
    nullable_object_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    force_string_putter_factory, optional_entity_id_array_optional_putter_factory,
    nullable_object_array_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_array_validator_factory, entity_id_validator_factory,
    force_string_validator_factory, nullable_object_array_validator_factory, preinstanced_validator_factory
)
from ...role import Role
from ...user import ClientUserBase

from ..action import AutoModerationAction
from ..trigger_metadata import AutoModerationRuleTriggerMetadataBase

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .preinstanced import AutoModerationEventType, AutoModerationRuleTriggerType


# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('rule_id')

# actions

parse_actions = nullable_object_array_parser_factory('actions', AutoModerationAction)
put_actions = nullable_object_array_optional_putter_factory('actions')
validate_actions = nullable_object_array_validator_factory('actions', AutoModerationAction)

# creator_id

parse_creator_id = entity_id_parser_factory('creator_id')
put_creator_id = entity_id_optional_putter_factory('creator_id')
validate_creator_id = entity_id_validator_factory('creator_id', ClientUserBase)

# enabled

parse_enabled = bool_parser_factory('enabled', True)
put_enabled = bool_optional_putter_factory('enabled', True)
validate_enabled = bool_validator_factory('enabled', True)

# event_type

parse_event_type = preinstanced_parser_factory('event_type', AutoModerationEventType, AutoModerationEventType.none)
put_event_type = preinstanced_putter_factory('event_type')
validate_event_type = preinstanced_validator_factory('event_type', AutoModerationEventType)

# excluded_channel_ids

parse_excluded_channel_ids = entity_id_array_parser_factory('exempt_channels')
put_excluded_channel_ids = optional_entity_id_array_optional_putter_factory('exempt_channels')
validate_excluded_channel_ids = entity_id_array_validator_factory('excluded_channel_ids', Channel)

# excluded_role_ids

parse_excluded_role_ids = entity_id_array_parser_factory('exempt_roles')
put_excluded_role_ids = optional_entity_id_array_optional_putter_factory('exempt_roles')
validate_excluded_role_ids = entity_id_array_validator_factory('excluded_role_ids', Role)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# trigger_type

parse_trigger_type = preinstanced_parser_factory(
    'trigger_type', AutoModerationRuleTriggerType, AutoModerationRuleTriggerType.none
)
put_trigger_type = preinstanced_putter_factory('trigger_type')
validate_trigger_type = preinstanced_validator_factory('trigger_type', AutoModerationRuleTriggerType)

# trigger_metadata

def parse_trigger_metadata(data, rule_trigger_type):
    """
    Parsers out an auto moderation rule's trigger metadata form the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Auto moderation rule data.
    
    rule_trigger_type : ``AutoModerationRuleTriggerType``
        The rule trigger's type.
    
    Returns
    -------
    trigger_metadata : ``AutoModerationRuleTriggerMetadataBase``
    """
    trigger_metadata_type = rule_trigger_type.metadata_type
    trigger_metadata_data = data.get('trigger_metadata', None)
    
    if trigger_metadata_data is None:
        trigger_metadata = trigger_metadata_type()
    else:
        trigger_metadata = trigger_metadata_type.from_data(trigger_metadata_data)
    
    return trigger_metadata


def put_trigger_metadata(trigger_metadata, data, defaults):
    """
    Puts the given rule trigger metadata's data into the given `data` json serializable object.
    
    Parameters
    ----------
    trigger_metadata : ``AutoModerationRuleTriggerMetadataBase``
        Rule trigger metadata.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (type(trigger_metadata) is not AutoModerationRuleTriggerMetadataBase):
        data['trigger_metadata'] = trigger_metadata.to_data(defaults = defaults)
    
    return data



__all__ = ('AuditLogEntryDetailConversion', )

from scarletio import RichAttributeErrorBaseType

from ..conversion_helpers.helpers import _eq_functions, _hash_function


class AuditLogEntryDetailConversion(RichAttributeErrorBaseType):
    """
    Represents how details are converted of an audit log entry.
    
    Attributes
    ----------
    field_key : `str`
        The serialised name of the field.
    field_name : `str`
        The field's name.
    value_deserializer : `FunctionType | MethodType`
        Raw to processed value converter.
    value_serializer : `FunctionType | MethodType`
        Processed to raw value converter.
    value_validator : `FunctionType | MethodType`
        Detail value validator.
    """
    __slots__ = ('field_key', 'field_name', 'value_deserializer', 'value_serializer', 'value_validator')
    
    def __new__(cls, field_key, field_name, *, value_deserializer = ..., value_serializer = ..., value_validator = ...):
        """
        Creates a new detail conversion representation with the given fields.
        
        Parameters
        ----------
        field_key : `str`
            The serialised name of the field.
        field_name : `str`
            The field's name.
        value_deserializer : `FunctionType | MethodType`, Optional (Keyword only)
            Raw to processed value converter.
        value_serializer : `FunctionType | MethodType`, Optional (Keyword only)
            Processed to raw value converter.
        value_validator : `FunctionType | MethodType`, Optional (Keyword only)
            Detail value validator.
        """
        if value_deserializer is ...:
            value_deserializer = None
        
        if value_serializer is ...:
            value_serializer = None
        
        if value_validator is ...:
            value_validator = None
        
        self = object.__new__(cls)
        self.field_key = field_key
        self.field_name = field_name
        self.value_deserializer = value_deserializer
        self.value_serializer = value_serializer
        self.value_validator = value_validator
        return self
    
    
    def __repr__(self):
        """Returns the detail conversion's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # field_key
        repr_parts.append(' field_key = ')
        repr_parts.append(repr(self.field_key))
        
        # field_name
        repr_parts.append(', field_name = ')
        repr_parts.append(repr(self.field_name))
        
        # value_deserializer
        value_deserializer = self.value_deserializer
        if value_deserializer is not None:
            repr_parts.append(', value_deserializer = ')
            repr_parts.append(value_deserializer.__module__)
            repr_parts.append('.')
            repr_parts.append(value_deserializer.__name__)
        
        # value_serializer
        value_serializer = self.value_serializer
        if value_serializer is not None:
            repr_parts.append(', value_serializer = ')
            repr_parts.append(value_serializer.__module__)
            repr_parts.append('.')
            repr_parts.append(value_serializer.__name__)
        
        # value_validator
        value_validator = self.value_validator
        if value_validator is not None:
            repr_parts.append(', value_validator = ')
            repr_parts.append(value_validator.__module__)
            repr_parts.append('.')
            repr_parts.append(value_validator.__name__)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the entry detail conversion's hash value."""
        hash_value = 0
        
        # field_key
        field_key = self.field_key
        hash_value ^= hash(field_key)
        
        # field_name
        field_name = self.field_name
        if field_key != field_name:
            hash_value ^= hash(field_name)
        
        # value_deserializer
        value_deserializer = self.value_deserializer
        if value_deserializer is not None:
            hash_value ^= _hash_function(value_deserializer)
        
        # value_serializer
        value_serializer = self.value_serializer
        if value_serializer is not None:
            hash_value ^= _hash_function(value_serializer)
        
        # value_validator
        value_validator = self.value_validator
        if value_validator is not None:
            hash_value ^= _hash_function(value_validator)
        
        return hash_value
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # field_key
        if self.field_key != other.field_key:
            return False
        
        # field_name
        if self.field_name != other.field_name:
            return False
        
        # value_deserializer
        if not _eq_functions(self.value_deserializer, other.value_deserializer):
            return False
        
        # value_serializer
        if not _eq_functions(self.value_serializer, other.value_serializer):
            return False
        
        # value_validator
        if not _eq_functions(self.value_validator, other.value_validator):
            return False
        
        return True
    
    
    def set_value_deserializer(self, value_deserializer):
        """
        Sets the get converter of the conversion and returns it back.
        
        Parameters
        ----------
        value_deserializer : `FunctionType | MethodType`
            Converter to put.
        
        Returns
        -------
        value_deserializer : `value_deserializer`
        """
        self.value_deserializer = value_deserializer
        return value_deserializer
    
    
    def set_value_serializer(self, value_serializer):
        """
        Sets the put converter of the conversion and returns it back.
        
        Parameters
        ----------
        value_serializer : `FunctionType | MethodType`
            Converter to put.
        
        Returns
        -------
        value_serializer : `value_serializer`
        """
        self.value_serializer = value_serializer
        return value_serializer
    
    
    def set_value_validator(self, value_validator):
        """
        Sets the converter of the conversion and returns it back.
        
        Parameters
        ----------
        value_validator : `FunctionType | MethodType`
            Converter to put.
        
        Returns
        -------
        value_validator : `value_validator`
        """
        self.value_validator = value_validator
        return value_validator



__all__ = ()

from .....discord.application import ApplicationIntegrationType
from .....discord.guild import Guild
from .....discord.application_command.application_command.constants import (
    NAME_LENGTH_MAX as APPLICATION_COMMAND_NAME_LENGTH_MAX,
    NAME_LENGTH_MIN as APPLICATION_COMMAND_NAME_LENGTH_MIN
)
from .....discord.application_command.application_command.preinstanced import (
    ApplicationCommandIntegrationContextType,
)
from .....discord.permission import Permission
from .....discord.preconverters import preconvert_bool, preconvert_flag, preconvert_snowflake



def _reset_application_command_schema(entity):
    """
    Resets the slasher application commands schema.
    
    Parameters
    ----------
    entity : ``CommandBaseApplicationCommand``
        The command to reset's its schema if applicable.
    """
    schema = entity._schema
    if (schema is not None):
        entity._schema = None
        
        parent_reference = entity._parent_reference
        if (parent_reference is not None):
            slasher = parent_reference()
            
            if (slasher is not None):
                slasher._schema_reset(entity)


def _validate_delete_on_unload(delete_on_unload):
    """
    Validates the given `delete_on_unload` value.
    
    Parameters
    ----------
    delete_on_unload : `bool`
        The value to validate.
    
    Returns
    -------
    delete_on_unload : `bool`
    
    Raises
    ------
    TypeError
        If `delete_on_unload` type is incorrect.
    """
    return preconvert_bool(delete_on_unload, 'delete_on_unload')


def _validate_nsfw(nsfw):
    """
    Validates the given `nsfw` value.
    
    Parameters
    ----------
    nsfw : `bool`
        The value to validate.
    
    Returns
    -------
    nsfw : `bool`
    
    Raises
    ------
    TypeError
        If `nsfw` type is incorrect.
    """
    return preconvert_bool(nsfw, 'nsfw')


def _validate_is_global(is_global):
    """
    Validates the given `is_global` value.
    
    Parameters
    ----------
    is_global : `bool`
        The value to validate.
    
    Returns
    -------
    is_global : `bool`
    
    Raises
    ------
    TypeError
        If `is_global` type is incorrect.
    """
    return preconvert_bool(is_global, 'is_global')


def _validate_1_guild(guild):
    """
    Validates 1 guild value.
    
    Parameters
    ----------
    guild : `Guild | int`
        The guild value to validate.
    
    Returns
    -------
    guild_id : `int`
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    ValueError
        If `guild`'s value is incorrect.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
    elif isinstance(guild, (int, str)):
        guild_id = preconvert_snowflake(guild, 'guild')
    else:
        raise TypeError(
            f'`guild` can be `{type(Guild).__name__}`, `int`, got {type(guild).__name__}; {guild!r}.'
        )
    
    return guild_id


def _validate_guild(guild):
    """
    Validates the given `guild` value.
    
    Parameters
    ----------
    guild : `Guild | int | (list | set)<Guild | int>`
        The `is_global` value to validate.
    
    Returns
    -------
    guild_ids : `set<int>`
        The validated `guild` value.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    ValueError
        If `guild`'s value is incorrect.
    """
    guild_ids = set()
    
    if isinstance(guild, (list, set)):
        for guild_value in guild:
            guild_id = _validate_1_guild(guild_value)
            guild_ids.add(guild_id)
        
        if not guild_ids:
            raise ValueError(
                f'`guild` cannot be empty, got {guild!r}.'
            )
    else:
        guild_id = _validate_1_guild(guild)
        guild_ids.add(guild_id)
    
    return guild_ids


def _validate_name(name):
    """
    Validates the given name.
    
    Parameters
    ----------
    name : `None`, `str`
        A command's respective name.
    
    Returns
    -------
    name : `None`, `str`
        The validated name.
    
    Raises
    ------
    TypeError
        If `name` is not given as `None` neither as `str`.
    ValueError
        If `name` length is out of the expected range [1:32].
    """
    if name is not None:
        name_type = type(name)
        if name_type is str:
            pass
        elif issubclass(name_type, str):
            name = str(name)
        else:
            raise TypeError(
                f'`name` can be `None`, `str`, got {name_type.__name__}; {name!r}.'
            )
        
        name_length = len(name)
        if (
            name_length < APPLICATION_COMMAND_NAME_LENGTH_MIN or
            name_length > APPLICATION_COMMAND_NAME_LENGTH_MAX
        ):
            raise ValueError(
                f'`name` length is out of the expected range '
                f'[{APPLICATION_COMMAND_NAME_LENGTH_MIN}:'
                f'{APPLICATION_COMMAND_NAME_LENGTH_MAX}], got {name_length!r}; {name!r}.'
            )
    
    return name


def _validate_required_permissions(required_permissions):
    """
    Validates the given `required_permissions` value.
    
    Parameters
    ----------
    required_permissions : `int | Permission`
        The `required_permissions` value to validate.
    
    Returns
    -------
    required_permissions : ``Permission``
    
    Raises
    ------
    TypeError
        If `required_permissions`'s type is incorrect.
    """
    return preconvert_flag(required_permissions, 'required_permissions', Permission)


APPLICATION_COMMAND_INTEGRATION_CONTEXT_TYPES_BY_ATTRIBUTE_NAME = {
    integration_context_type.name.replace(' ', '_'): integration_context_type
    for integration_context_type
    in ApplicationCommandIntegrationContextType.INSTANCES.values()
}


APPLICATION_INTEGRATION_TYPES_BY_ATTRIBUTE_NAME = {
    integration_type.name.replace(' ', '_'): integration_type
    for integration_type
    in ApplicationIntegrationType.INSTANCES.values()
    
}


def _fail_preinstanced_array_validation_single(field_name, preinstanced_type, string_resolution_table, value):
    """
    Fails preinstanced array validation when checking for a single instance.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    preinstanced_type : `type<PreinstancedBase>`
        The preinstanced type we are validating for.
    string_resolution_table : `dict<str, PreinstancedBase>`
        String resolution table for the preinstanced type.
    value : `object`
        The value that failed on validation.
    
    Raises
    ------
    TypeError
    """
    allowed_strings_string = ', '.join(sorted(repr(string) for string in string_resolution_table.keys()))
    
    raise TypeError(
        f'`{field_name}` can be `None`, `{preinstanced_type.__name__}`, `int`, '
        f'`str` (any of: {allowed_strings_string!s}), or `iterable`, got '
        f'{type(value).__name__}; {value!r}.'
    )


def _fail_preinstanced_array_validation_element(
    field_name, preinstanced_type, string_resolution_table, preinstanced_array, value
):
    """
    Fails preinstanced array validation when checking for an element.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    preinstanced_type : `type<PreinstancedBase>`
        The preinstanced type we are validating for.
    string_resolution_table : `dict<str, PreinstancedBase>`
        String resolution table for the preinstanced type.
    preinstanced_array : `object`
        The given value to validate.
    value : `object`
        The value that failed on validation.
    
    Raises
    ------
    TypeError
    """
    allowed_strings_string = ', '.join(sorted(repr(string) for string in string_resolution_table.keys()))
    
    raise TypeError(
        f'`{field_name}` elements can be `{preinstanced_type.__name__}`, `int`, '
        f'`str` (any of: {allowed_strings_string!s}), or `iterable`, got '
        f'got {type(value).__name__}; {value!r}; preinstanced_array = {preinstanced_array!r}'
    )


def _pre_validate_preinstanced_array(field_name, preinstanced_type, string_resolution_table, preinstanced_array):
    """
    Validates the given preinstanced array.
    
    Parameters
    ----------
    field_name : `str`
        The field's name.
    preinstanced_type : `type<PreinstancedBase>`
        The preinstanced type we are validating for.
    string_resolution_table : `dict<str, preinstanced_type>`
        String resolution table for the preinstanced type.
    preinstanced_array : `object`
        The given value to validate.
    
    Returns
    -------
    preinstanced_array : `None | tuple<preinstanced_type>`
    
    Raises
    ------
    TypeError
    """
    if preinstanced_array is None:
        return None
    
    if isinstance(preinstanced_array, preinstanced_type):
        return (preinstanced_array,)
    
    if isinstance(preinstanced_array, str):
        preinstanced = string_resolution_table.get(preinstanced_array, None)
        if preinstanced is None:
            _fail_preinstanced_array_validation_single(
                field_name, preinstanced_type, string_resolution_table, preinstanced_array
            )
        
        return (preinstanced,)
    
    if isinstance(preinstanced_array, int):
        return (ApplicationCommandIntegrationContextType(preinstanced_array),)
    
    if getattr(preinstanced_array, '__iter__', None) is None:
        _fail_preinstanced_array_validation_single(
            field_name, preinstanced_type, string_resolution_table, preinstanced_array
        )
    
    unique_elements = None
    
    for preinstanced in preinstanced_array:
        if isinstance(preinstanced, preinstanced_type):
            pass
        
        elif isinstance(preinstanced, str):
            preinstanced = string_resolution_table.get(preinstanced, None)
            if preinstanced is None:
                _fail_preinstanced_array_validation_element(
                    field_name, preinstanced_type, string_resolution_table, preinstanced_array, preinstanced
                )
            
        elif isinstance(preinstanced, int):
            preinstanced = preinstanced_type(preinstanced)
        
        else:
            _fail_preinstanced_array_validation_element(
                field_name, preinstanced_type, string_resolution_table, preinstanced_array, preinstanced
            )
        
        if unique_elements is None:
            unique_elements = set()
        
        unique_elements.add(preinstanced)
    
    if unique_elements is None:
        return None
    
    return tuple(sorted(unique_elements))


def _validate_integration_context_types(integration_context_types):
    """
    Validates the given `integration_context_types` value.
    
    Parameters
    ----------
    integration_context_types : `None`, ``ApplicationCommandIntegrationContextType``, `int`, `str`, \
            `iterable<ApplicationCommandIntegrationContextType | int | str>`
        The places where the application command shows up.
    
    Returns
    -------
    integration_context_types : `None | tuple<ApplicationCommandIntegrationContextType>`
    
    Raises
    ------
    TypeError
    """
    return _pre_validate_preinstanced_array(
        'integration_context_types',
        ApplicationCommandIntegrationContextType,
        APPLICATION_COMMAND_INTEGRATION_CONTEXT_TYPES_BY_ATTRIBUTE_NAME,
        integration_context_types,
    )


def _validate_integration_types(integration_types):
    """
    Validates the given `integration_types` value.
    
    Parameters
    ----------
    integration_types : `None`, ``ApplicationIntegrationType``, `int`, `str`, \
            `iterable<ApplicationIntegrationType | int | str>`
        The options where the application command can be integrated to.
    
    Returns
    -------
    integration_types : `None | tuple<ApplicationIntegrationType>`
    
    Raises
    ------
    TypeError
    """
    integration_types = _pre_validate_preinstanced_array(
        'integration_types',
        ApplicationIntegrationType,
        APPLICATION_INTEGRATION_TYPES_BY_ATTRIBUTE_NAME,
        integration_types,
    )
    if (integration_types is None):
        integration_types = (ApplicationIntegrationType.guild_install,)
    
    return integration_types



__all__ = ('ReactionType',)

from ...bases import Preinstance as P, PreinstancedBase


class ReactionType(PreinstancedBase, value_type = int):
    """
    Represents a reaction's type.
    
    Attributes
    ----------
    name : `str`
        The default name of the reaction type.
    
    value : `int`
        The discord side identifier value of the reaction type.
    
    Type Attributes
    ---------------
    Every predefined reaction type can be accessed as type attribute as well:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | name          | value |
    +======================+================+=======+
    | standard              | standard      | 0     |
    +-----------------------+---------------+-------+
    | burst                 | status        | 1     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    standard = P(0, 'standard')
    burst = P(1, 'burst')



__all__ = ()

from ...channel import Channel
from ...emoji import Emoji, create_partial_emoji_from_inline_data, put_partial_emoji_inline_data_into
from ...field_parsers import entity_id_parser_factory, nullable_string_parser_factory
from ...field_putters import (
    entity_id_putter_factory, nullable_functional_optional_putter_factory, nullable_string_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, nullable_entity_validator_factory, nullable_string_validator_factory
)

from .constants import DESCRIPTION_LENGTH_MAX

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# emoji

parse_emoji = create_partial_emoji_from_inline_data


def put_emoji(emoji, data, defaults):
    """
    Puts the emoji into the given `data` json serializable object.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The emoji.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (emoji is not None):
        put_partial_emoji_inline_data_into(emoji, data)
    
    return data


validate_emoji = nullable_entity_validator_factory('emoji', Emoji)



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, field_parser_factory, force_string_parser_factory, nullable_object_array_parser_factory,
    preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, force_string_putter_factory, nullable_field_optional_putter_factory,
    nullable_object_array_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, field_validator_factory, force_string_validator_factory,
    nullable_object_array_validator_factory, preinstanced_validator_factory
)

from ...application_command import ApplicationCommandOptionType
from ...application_command.application_command_option.constants import (
    NAME_LENGTH_MAX as APPLICATION_COMMAND_NAME_LENGTH_MAX, NAME_LENGTH_MIN as APPLICATION_COMMAND_NAME_LENGTH_MIN
)


# focused

parse_focused = bool_parser_factory('focused', False)
put_focused = bool_optional_putter_factory('focused', False)
validate_focused = bool_validator_factory('focused', False)

# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory(
    'name', APPLICATION_COMMAND_NAME_LENGTH_MIN, APPLICATION_COMMAND_NAME_LENGTH_MAX
)

# options

parse_options = nullable_object_array_parser_factory('options', NotImplemented, include = 'InteractionOption')
put_options = nullable_object_array_optional_putter_factory('options')
validate_options = nullable_object_array_validator_factory('options', NotImplemented, include = 'InteractionOption')

# type

parse_type = preinstanced_parser_factory('type', ApplicationCommandOptionType, ApplicationCommandOptionType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', ApplicationCommandOptionType)

# value

parse_value = field_parser_factory('value')
put_value = nullable_field_optional_putter_factory('value')
validate_value = field_validator_factory('value')



__all__ = ()

from random import randint

from scarletio import copy_docs

from ..packets.constants import MAX_UINT_32
from ..packets.utils import create_rtp_data_lite, create_rtp_header_lite

from .base import EncryptionAdapterBase

try:
    from libnacl import (
        crypto_aead_aes256gcm_KEYBYTES, crypto_aead_aes256gcm_NPUBBYTES, crypto_aead_aes256gcm_decrypt,
        crypto_aead_aes256gcm_encrypt
    )
except ImportError:
    crypto_aead_aes256gcm_KEYBYTES = 0
    crypto_aead_aes256gcm_NPUBBYTES = 0
    crypto_aead_aes256gcm_encrypt = lambda cipher_data, additional_authentication_data, nonce, key: b''
    crypto_aead_aes256gcm_decrypt = lambda plain_data, additional_authentication_data, nonce, key: b''


class EncryptionAdapter__aead_aes256_gcm_rtpsize(EncryptionAdapterBase):
    """
    `aead_aes256_gcm_rtpsize` encryption adapter.
    
    Attributes
    ----------
    key : `bytes`
        Secret key used for encrypting / decrypting.
    
    nonce_counter : `int`
        Incremental counter for generating nonces.
    """
    __slots__ = ('nonce_counter',)
    
    available = True if crypto_aead_aes256gcm_KEYBYTES else False
    key_length = crypto_aead_aes256gcm_KEYBYTES
    name = 'aead_aes256_gcm_rtpsize'
    nonce_length = crypto_aead_aes256gcm_NPUBBYTES
    priority = 4
    
    
    @copy_docs(EncryptionAdapterBase.__new__)
    def __new__(cls, key):
        self = object.__new__(cls)
        self.key = key
        self.nonce_counter = randint(0, MAX_UINT_32) % 487
        return self
    
    
    def get_next_nonce_bytes(self):
        """
        Increments the nonce counter and return a new nonce value generated from it.
        
        Returns
        -------
        nonce : `bytes`
        """
        nonce_counter = self.nonce_counter + 1
        self.nonce_counter = nonce_counter
        return (nonce_counter & MAX_UINT_32).to_bytes(4, 'big')
    
    
    @copy_docs(EncryptionAdapterBase.create_send_packet)
    def create_send_packet(self, voice_client, plain_data):
        header = create_rtp_header_lite(
            voice_client._sequence, voice_client._timestamp, voice_client._audio_source
        )
        nonce_bytes = self.get_next_nonce_bytes()
        nonce = nonce_bytes.ljust(self.nonce_length, b'\x00')
        payload = crypto_aead_aes256gcm_encrypt(plain_data, header, nonce, self.key)
        return create_rtp_data_lite(header, payload, nonce_bytes)
    
    
    @copy_docs(EncryptionAdapterBase.process_received_payload)
    def process_received_payload(self, rtp_packet):
        # last 4 bits of payload is the nonce rest is the actual payload
        payload = rtp_packet.payload
        nonce = bytes(payload[-4 :]).ljust(self.nonce_length, b'\x00')
        payload = bytes(payload[: -4])
        
        return crypto_aead_aes256gcm_decrypt(payload, bytes(rtp_packet.header), nonce, self.key)
